import typing
from enum import Enum
from itp_interface.rl.proof_action import ProofAction
from itp_interface.tools.training_data_format import TrainingDataFormat
try:
    from .grammars.grammar import Grammar
except ImportError:
    from grammars.grammar import Grammar

from dataclasses import dataclass, field
from dataclasses_json import dataclass_json

class CoqGptResponseActions(object):
    GOALS = "[GOALS]"
    ERROR = "[ERROR]"
    ERROR_MESSAGE = "[ERROR MESSAGE]"

@dataclass_json
@dataclass
class CoqGptResponse(object):
    action : str = CoqGptResponseActions.GOALS
    success: bool = True
    message: str = ""
    steps: typing.List[str] = field(default_factory=list)
    incorrect_steps: typing.List[str] = field(default_factory=list)
    error_message: typing.Optional[str] = None
    last_step: typing.Optional[str] = None
    training_data_format: typing.Optional[TrainingDataFormat] = None
    informal_proof: typing.Optional[str] = None
    informal_theorem: typing.Optional[str] = None

class CoqGPTResponseDfsGrammar(Grammar):
    grammar = """
Prog:
  GoalsResponse
| ErrorResponse
| String Prog
| Prog String Prog;
ErrorResponse:
   Error String End
|  Error ErrorString End;
GoalsResponse:
  Goals Description String GoalResponses StepsResponses IncorrectStepsResponses LastResponse End
| Goals GoalResponses StepsResponses IncorrectStepsResponses LastResponse End;
GoalResponses:
  GoalResponse
| GoalResponse GoalResponses
| EMPTY;
GoalResponse:
 Goal int String HypsResponses DfnsResponses ThmsResponses;
DfnsResponses:
    Dfns int DfnResponses
|   EMPTY;
DfnResponses:
    DfnResponse
|   DfnResponse DfnResponses;
DfnResponse:
    Dfn String;
ThmsResponses:
    Thms int ThmResponses
|   EMPTY;
ThmResponses:
    ThmResponse
|   ThmResponse ThmResponses;
ThmResponse:
    Thm String;
HypsResponses:
    Hyps int HypResponses
|   EMPTY;
HypResponses:
  HypResponse
| HypResponse HypResponses;
HypResponse:
  Hyp String;
IncorrectStepsResponses:
    IncrctStps StepResponses
|   EMPTY;
LastResponse:
    LastStep String ErrorMessage String
|   LastStep String Success
|   EMPTY;
StepsResponses:
    Stps StepResponses
|   EMPTY;
StepResponses:
    StepResponse
|   StepResponse StepResponses;
StepResponse:
    Stp String;


terminals
int: /\d+/;
Goals: "[GOALS]";
Goal: "[GOAL]";
Hyps: "[HYPOTHESES]";
Hyp: "[HYPOTHESIS]";
Stps: "[STEPS]";
Stp: "[STEP]";
IncrctStps: "[INCORRECT STEPS]";
Dfns: "[DEFINITIONS]";
Dfn: "[DEFINITION]";
Thms: "[THEOREMS]";
Thm: "[THEOREM]";
Error: "[ERROR]";
ErrorMessage: "[ERROR MESSAGE]";
Success: "[SUCCESS]";
LastStep: "[LAST STEP]";
End: "[END]";
Description: "[DESCRIPTION]";
String:;
ErrorString:;
"""
    class Keywords(Enum):
        GOALS = "[GOALS]"
        GOAL = "[GOAL]"
        HYPOTHESES = "[HYPOTHESES]"
        HYPOTHESIS = "[HYPOTHESIS]"
        STEPS = "[STEPS]"
        STEP = "[STEP]"
        INCORRECT_STEPS = "[INCORRECT STEPS]"
        DEFINITIONS = "[DEFINITIONS]"
        DEFINITION = "[DEFINITION]"
        THEOREMS = "[THEOREMS]"
        THEOREM = "[THEOREM]"
        ERROR = "[ERROR]"
        ERROR_MESSAGE = "[ERROR MESSAGE]"
        SUCCESS = "[SUCCESS]"
        END = "[END]"
        DESCRIPTION = "[DESCRIPTION]"
        LAST_STEP = "[LAST STEP]"
        INFORMAL_PROOF = "[INFORMAL-PROOF]"
        INFORMAL_THEOREM = "[INFORMAL-THEOREM]"

        def __str__(self) -> str:
            return self.value

    keywords = [keyword.value for keyword in Keywords]

    def before_keyword(text, pos):
        last = pos
        while last < len(text):
          while last < len(text) and text[last] != '[':
            last += 1
          if last < len(text):
              for keyword in CoqGPTResponseDfsGrammar.keywords:
                  if text[last:].startswith(keyword):
                      return text[pos:last]
              last += 1
    
    def error_string(text, pos):
        last = pos
        while last < len(text):
            while last < len(text) and text[last] != '[':
                last += 1
            if last < len(text) and text[last:].startswith("[END]") and text[last:].endswith("[END]"):
                return text[pos:last]
            last += 1

    def __init__(self):
        recognizers = {
            'String': CoqGPTResponseDfsGrammar.before_keyword,
            'ErrorString': CoqGPTResponseDfsGrammar.error_string
        }
        super(CoqGPTResponseDfsGrammar, self).__init__(CoqGPTResponseDfsGrammar.grammar, CoqGPTResponseDfsGrammar.keywords, recognizers=recognizers)
    
    def format_as_per_grammar(self, coq_gpt_response: CoqGptResponse, k: typing.Optional[int] = None, max_token_cnt: typing.Optional[int] = None, characters_per_token: float = 4.0) -> str:
        # Add algorithm for trimming the right amount of goals, theorems and defintions, steps, etc. based on the max_token_cnt
        char_cnt = int(max_token_cnt * characters_per_token) if max_token_cnt is not None else None # 4 is the average length of a token as per OpenAI
        text = ""
        if coq_gpt_response.action == CoqGptResponseActions.ERROR:
            text = f"{CoqGPTResponseDfsGrammar.Keywords.ERROR}\n{coq_gpt_response.message}\n{CoqGPTResponseDfsGrammar.Keywords.END}"
        elif coq_gpt_response.action == CoqGptResponseActions.GOALS:
            lines_map = {
                CoqGPTResponseDfsGrammar.Keywords.GOALS : [],
                CoqGPTResponseDfsGrammar.Keywords.INFORMAL_THEOREM : [],
                CoqGPTResponseDfsGrammar.Keywords.INFORMAL_PROOF : [],
                CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS : [],
                CoqGPTResponseDfsGrammar.Keywords.THEOREMS : [],
                CoqGPTResponseDfsGrammar.Keywords.STEPS : [],
                CoqGPTResponseDfsGrammar.Keywords.INCORRECT_STEPS : [],
                CoqGPTResponseDfsGrammar.Keywords.LAST_STEP : [],
                CoqGPTResponseDfsGrammar.Keywords.SUCCESS : [],
                CoqGPTResponseDfsGrammar.Keywords.ERROR_MESSAGE : []
            }
            lines_order = [
                CoqGPTResponseDfsGrammar.Keywords.GOALS,
                CoqGPTResponseDfsGrammar.Keywords.INFORMAL_THEOREM,
                CoqGPTResponseDfsGrammar.Keywords.INFORMAL_PROOF,
                CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS,
                CoqGPTResponseDfsGrammar.Keywords.THEOREMS,
                CoqGPTResponseDfsGrammar.Keywords.STEPS,
                CoqGPTResponseDfsGrammar.Keywords.INCORRECT_STEPS,
                CoqGPTResponseDfsGrammar.Keywords.LAST_STEP,
                CoqGPTResponseDfsGrammar.Keywords.SUCCESS,
                CoqGPTResponseDfsGrammar.Keywords.ERROR_MESSAGE
            ]
            priority_order_lo_hi = [
                CoqGPTResponseDfsGrammar.Keywords.THEOREMS,
                CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS,
                CoqGPTResponseDfsGrammar.Keywords.STEPS,
                CoqGPTResponseDfsGrammar.Keywords.INFORMAL_PROOF,
                CoqGPTResponseDfsGrammar.Keywords.INFORMAL_THEOREM,
                CoqGPTResponseDfsGrammar.Keywords.GOALS, # trim down the goals
                CoqGPTResponseDfsGrammar.Keywords.INCORRECT_STEPS,
                CoqGPTResponseDfsGrammar.Keywords.ERROR_MESSAGE,
                CoqGPTResponseDfsGrammar.Keywords.LAST_STEP,
                CoqGPTResponseDfsGrammar.Keywords.SUCCESS,
            ]
            assert coq_gpt_response.training_data_format is not None
            new_line = f"Goals to prove:\n{CoqGPTResponseDfsGrammar.Keywords.GOALS}"
            lines_map[CoqGPTResponseDfsGrammar.Keywords.GOALS] = [new_line]
            if coq_gpt_response.training_data_format.goal_description is not None:
                new_line = f"{CoqGPTResponseDfsGrammar.Keywords.DESCRIPTION}\n{coq_gpt_response.training_data_format.goal_description}\n"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.GOALS].append(new_line)
            for i, goal in enumerate(coq_gpt_response.training_data_format.start_goals):
                new_line = f"{CoqGPTResponseDfsGrammar.Keywords.GOAL} {i+1}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.GOALS].append(new_line)
                new_line = str(goal.goal)
                lines_map[CoqGPTResponseDfsGrammar.Keywords.GOALS].append(new_line)
                if len(goal.hypotheses) > 0:
                    new_line = f"{CoqGPTResponseDfsGrammar.Keywords.HYPOTHESES} {i + 1}"
                    lines_map[CoqGPTResponseDfsGrammar.Keywords.GOALS].append(new_line)
                    for hyp in goal.hypotheses:
                        new_line = f"{CoqGPTResponseDfsGrammar.Keywords.HYPOTHESIS} {hyp}"
                        lines_map[CoqGPTResponseDfsGrammar.Keywords.GOALS].append(new_line)
                if len(goal.relevant_defns) > 0 and (k is None or k > 0):
                    dfns = goal.relevant_defns
                    if k is not None:
                        dfns = dfns[:k]
                    dfns = [str(coq_gpt_response.training_data_format.all_useful_defns_theorems[dfn.lemma_idx]) for dfn in dfns]
                    new_line = f"{CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS} {i + 1}"
                    if len(lines_map[CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS]) == 0:
                        lines_map[CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS] = [new_line]
                    else:
                        lines_map[CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS].append(new_line)
                    lines_map[CoqGPTResponseDfsGrammar.Keywords.DEFINITIONS].extend([f"{CoqGPTResponseDfsGrammar.Keywords.DEFINITION} {dfn}" for dfn in dfns])
                if len(goal.possible_useful_theorems_external) + len(goal.possible_useful_theorems_local) > 0 and (k is None or k > 0):
                    thms = goal.possible_useful_theorems_local + goal.possible_useful_theorems_external
                    if k is not None:
                        thms = thms[:k]
                    thms = [str(coq_gpt_response.training_data_format.all_useful_defns_theorems[thm.lemma_idx]) for thm in thms]
                    new_line = f"{CoqGPTResponseDfsGrammar.Keywords.THEOREMS} {i + 1}"
                    if len(lines_map[CoqGPTResponseDfsGrammar.Keywords.THEOREMS]) == 0:
                        lines_map[CoqGPTResponseDfsGrammar.Keywords.THEOREMS] = [new_line]
                    else:
                        lines_map[CoqGPTResponseDfsGrammar.Keywords.THEOREMS].append(new_line)
                    lines_map[CoqGPTResponseDfsGrammar.Keywords.THEOREMS].extend([f"{CoqGPTResponseDfsGrammar.Keywords.THEOREM} {thm}" for thm in thms])
            if len(coq_gpt_response.steps) > 0:
                new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.STEPS}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.STEPS] = [new_line]
                lines_map[CoqGPTResponseDfsGrammar.Keywords.STEPS].extend([f"{CoqGPTResponseDfsGrammar.Keywords.STEP} {step}" for step in coq_gpt_response.steps])
            if len(coq_gpt_response.incorrect_steps) > 0:
                new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.INCORRECT_STEPS}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.INCORRECT_STEPS] = [new_line]
                lines_map[CoqGPTResponseDfsGrammar.Keywords.INCORRECT_STEPS].extend([f"{CoqGPTResponseDfsGrammar.Keywords.STEP} {step}" for step in coq_gpt_response.incorrect_steps])
            if coq_gpt_response.last_step is not None:
                new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.LAST_STEP}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.LAST_STEP] = [new_line]
                lines_map[CoqGPTResponseDfsGrammar.Keywords.LAST_STEP].append(coq_gpt_response.last_step)
                if coq_gpt_response.success:
                    new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.SUCCESS}"
                    lines_map[CoqGPTResponseDfsGrammar.Keywords.SUCCESS] = [new_line]
            if coq_gpt_response.error_message is not None:
                # assert coq_gpt_response.last_step is not None
                # assert not coq_gpt_response.success
                new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.ERROR_MESSAGE}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.ERROR_MESSAGE] = [new_line]
                lines_map[CoqGPTResponseDfsGrammar.Keywords.ERROR_MESSAGE].append(coq_gpt_response.error_message)
            if coq_gpt_response.informal_theorem is not None:
                new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.INFORMAL_THEOREM}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.INFORMAL_THEOREM] = [new_line]
                lines_map[CoqGPTResponseDfsGrammar.Keywords.INFORMAL_THEOREM].append(coq_gpt_response.informal_theorem)
            if coq_gpt_response.informal_proof is not None:
                new_line = f"\n{CoqGPTResponseDfsGrammar.Keywords.INFORMAL_PROOF}"
                lines_map[CoqGPTResponseDfsGrammar.Keywords.INFORMAL_PROOF] = [new_line]
                lines_map[CoqGPTResponseDfsGrammar.Keywords.INFORMAL_PROOF].append(coq_gpt_response.informal_proof)
            keywords = [keyword for keyword in lines_map.keys()]
            # Convert all the lines under each keyword to a single string
            for keyword in keywords:
                lines_map[keyword] = "\n".join(lines_map[keyword])
            # Frame the first prompt version without any token limit
            text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0]) + f"\n{CoqGPTResponseDfsGrammar.Keywords.END}"
            
            # Trim the lines based on the max_token_cnt
            if char_cnt is not None and len(text) > char_cnt:
                _idx = 0
                diff = len(text) - char_cnt
                while _idx < len(priority_order_lo_hi) and diff > 0:
                    trim_part = priority_order_lo_hi[_idx]
                    if trim_part in lines_map:
                        if trim_part == CoqGPTResponseDfsGrammar.Keywords.STEPS:
                            if len(lines_map[trim_part]) <= diff:
                                lines_map[trim_part] = ""
                            else:
                                lines_map[trim_part] = lines_map[trim_part][diff:]
                        else:
                            lines_map[trim_part] = lines_map[trim_part][:-diff] # Trim everything except the STEPS from the end
                    text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0]) + f"\n{CoqGPTResponseDfsGrammar.Keywords.END}"
                    diff = len(text) - char_cnt
                    _idx += 1
        else:
            raise Exception(f"Invalid action {coq_gpt_response.action}")
        # verify that the text is valid as per grammar by compiling it
        # self.compile(text)
        if char_cnt is not None:
            assert len(text) <= char_cnt, f"Text length {len(text)} is greater than the max token count {char_cnt}. Possibly too few characters per token." +\
            f" characters_per_token = {characters_per_token}, max_token_cnt = {max_token_cnt}"
            # text = text[:char_cnt] # Just trim the text from the end because no trimming strategy has worked out
        return text


class CoqGptRequestActions(object):
    RUN_TACTIC = "[RUN TACTIC]"
    GET_DFNS_THMS = "[GET DEFINITIONS AND THEOREMS]"

@dataclass_json
@dataclass
class CoqGptRequest(object):
    action : str = CoqGptRequestActions.RUN_TACTIC
    args: typing.List[str] = field(default_factory=list)

class CoqGPTRequestGrammar(Grammar):
    grammar = """
Prog: 
  RunTacticRequest
| GetDfnsThmsRequest
| String Prog;
GetDfnsThmsRequest:
    GetDfnsThms End;
RunTacticRequest:
    RunTactic StpRequests End;
StpRequests:
  String;

terminals
Stp: "[STEP]";
End: "[END]";
RunTactic: "[RUN TACTIC]";
GetDfnsThms: "[GET DEFINITIONS AND THEOREMS]";
String:;
"""
    keywords = ["[STEP]", "[END]", "[RUN TACTIC]", "[GET DEFINITIONS]"]

    end = "[END]"

    def before_keyword(text, pos):
        last = pos
        while last < len(text):
          while last < len(text) and text[last] != '[':
            last += 1
          if last < len(text):
              for keyword in CoqGPTRequestGrammar.keywords:
                  if text[last:].startswith(keyword):
                      return text[pos:last]
              last += 1

    def __init__(self, enable_defensive_parsing: bool = False):
        recognizers = {
            'String': CoqGPTRequestGrammar.before_keyword
        }
        self.enable_defensive_parsing = enable_defensive_parsing
        super(CoqGPTRequestGrammar, self).__init__(CoqGPTRequestGrammar.grammar, CoqGPTRequestGrammar.keywords, recognizers=recognizers)

    def _parse_expr(self, nonTerminal, nodes, context):
        if nonTerminal == "GetDfnsThmsRequest":
            context.action = CoqGptRequestActions.GET_DFNS_THMS
            context.args = [CoqGptRequestActions.GET_DFNS_THMS[1:-1]]
        elif nonTerminal == "RunTacticRequest":
            assert len(nodes) >= 2
            context.action = CoqGptRequestActions.RUN_TACTIC
            context.args.reverse()
        elif nonTerminal == "StpRequests":
            assert len(nodes) >= 1
            str_node = str(nodes[0]).strip()
            if len(str_node) > 0:
                context.args.append(str_node)
        else:
            raise Exception(f"Unknown non-terminal {nonTerminal}")
        return context
        
    def get_action(self, inp=None):
        context = CoqGptRequest()
        actions = {
            "Prog": lambda _, nodes: context,
            "GetDfnsThmsRequest": lambda _, nodes: self._parse_expr('GetDfnsThmsRequest', nodes, context),
            "RunTacticRequest": lambda _, nodes: self._parse_expr('RunTacticRequest', nodes, context),
            "StpRequests": lambda _, nodes: self._parse_expr('StpRequests', nodes, context),
            "String": lambda _, nodes: str(nodes) # Since this is always a string
        }
        return actions
    
    def interpret_result(self, result):
        assert isinstance(result, CoqGptRequest), f"Result must be a CoqGptRequest. Got {type(result)}"
        return result
    
    def generate_message_from_gpt_request(self, coq_gpt_request: CoqGptRequest) -> str:
        if coq_gpt_request.action == CoqGptRequestActions.RUN_TACTIC:
            args = '\n'.join(coq_gpt_request.args)
            return f"{CoqGptRequestActions.RUN_TACTIC}\n{args}\n{CoqGPTRequestGrammar.end}"
        elif coq_gpt_request.action == CoqGptRequestActions.GET_DFNS_THMS:
            return f"{CoqGptRequestActions.GET_DFNS_THMS}\n{CoqGPTRequestGrammar.end}"
        else:
            raise Exception(f"Invalid action {coq_gpt_request.action}")

    def get_openai_request(self, message_response: str) -> typing.Tuple[CoqGptRequest, str]:
        message, finish_reason = message_response
        defensive_parsing = finish_reason != "stop" or self.enable_defensive_parsing
        if defensive_parsing:     
            return self.defensive_parsing(message)
        else:
            return self.normal_parsing(message)
    
    def normal_parsing(self, message):
        message += CoqGPTRequestGrammar.end
        result : CoqGptRequest = self.run(message, None)            
        message = self.generate_message_from_gpt_request(result)
        return (result, message)

    def defensive_parsing(self, message):
        start_idx = 0
        end_idx = len(message)
        # Generate all possible sub-strings such that the start_idx is less than end_idx
        idxs = [(s_idx, e_idx) for s_idx in range(start_idx, end_idx) for e_idx in range(end_idx, s_idx, -1)]
        message_temp = message
        message_parsed = False
        for s_idx, e_idx in idxs:
            # This type of robust parsing can be needed in case of some LLMs which
            # don't follow the specified format
            try:
                message_temp = message[s_idx:e_idx]
                if message_temp.endswith(CoqGPTRequestGrammar.end):
                    # Just in case the LLM doesn't remove the stop token
                    message_temp = message_temp.strip(CoqGPTRequestGrammar.end)
                message_temp += f"\n{CoqGPTRequestGrammar.end}"
                result : CoqGptRequest = self.run(message_temp, None)            
                message_temp = self.generate_message_from_gpt_request(result)
                message_parsed = True
            except:
                message_parsed = False
            if message_parsed:
                break
        if not message_parsed:
            message_temp = message[start_idx:end_idx]
            message_temp += f"\n{CoqGPTRequestGrammar.end}"
            result : CoqGptRequest = self.run(message_temp, None)            
            message_temp = self.generate_message_from_gpt_request(result)
        return (result, message_temp)

    def attempt_parsing(self, message):
        # do a greedy correction to ensure that the message is parsable
        idx = len(message)
        exceptions = []
        message_seems_fixable = True
        try:
            # trim any unwanted keywords at the end
            idx = message.rfind('[')
            if idx < 0:
                raise Exception("No opening bracket found, message is not parsable")
            close_idx = message.rfind(']', idx, len(message))
            if close_idx < 0:
                message = message[:idx]
            else:
                idx = len(message)
        except Exception:
            message_seems_fixable = False
            pass
        if message_seems_fixable:    
            attempt = 0
            while idx >= 0:
                try:
                    parsable_message = message[:idx] + f"\n{CoqGPTRequestGrammar.end}"
                    self.compile(parsable_message)
                    break
                except Exception as e:
                    exceptions.append(e)
                    idx = message.rfind('[', 0, idx)
                attempt += 1
            if idx >= 0:
                message = parsable_message
            else:
                raise exceptions[0]
            result : CoqGptRequest = self.run(message, None)
            if result.action == CoqGptRequestActions.RUN_TACTIC and len(result.args) > 1:
                result.args = result.args[:-1] # remove the last tactic as it can be incomplete
        else:
            message += CoqGPTRequestGrammar.end
            result : CoqGptRequest = self.run(message, None)
        message = self.generate_message_from_gpt_request(result)
        return (result, message)

    def parse_request_to_args(self, messages: typing.List[str]) -> typing.List[str]:
        results : typing.List[str] = []
        for message in messages:
            assert message.endswith(CoqGPTRequestGrammar.end), "Message must end with end token"
            result : CoqGptRequest = self.run(message, None)
            results.extend(result.args)
        return results
    
class FewShotGptCoqKeywords(object):
    PROOF = "Proof."
    QED = "Qed."
    THEOREM = "[THEOREM]"
    DEFINITION = "[DEFINITION]"
    DEFINITIONS = "[DEFINITIONS]"
    LEMMAS = "[LEMMAS]"
    LEMMA = "[LEMMA]"
    END = "[END]"

class FewShotGptLeanKeywords(object):
    PROOF = "[PROOF]"
    QED = "[END]"
    THEOREM = "[THEOREM]"
    DEFINITION = "[DEFINITION]"
    DEFINITIONS = "[DEFINITIONS]"
    LEMMAS = "[LEMMAS]"
    LEMMA = "[LEMMA]"
    END = "[END]"
    INFORMAL_THEOREM = "[INFORMAL-THEOREM]"
    INFORMAL_PROOF = "[INFORMAL-PROOF]"

@dataclass_json
@dataclass
class FewShotGptRequest(object):
    action : ProofAction
    proof_string : str

@dataclass_json
@dataclass
class FewShotGptResponse(object):
    theorem: str
    defintions: typing.List[str] = field(default_factory=list)
    lemmas: typing.List[str] = field(default_factory=list)
    informal_theorem: typing.Optional[str] = None
    informal_proof: typing.Optional[str] = None

class FewShotGptResponseGrammar(Grammar):
    grammar = f"""
Prog:
  Thm String End
| Thm String DfnsResponses LmsResponses End
| Thm String InfThm String InfPrf String End;
DfnsResponses:
    Dfns DfnResponses
|   EMPTY;
DfnResponses:
    DfnResponse
|   DfnResponse DfnResponses;
DfnResponse:
    Dfn String;
LmsResponses:
    Lms LmResponses
|   EMPTY;
LmResponses:
    LmResponse
|   LmResponse LmResponses;
LmResponse:
    Lm String;
"""

    def before_keyword(self, text, pos):
        last = pos
        while last < len(text):
            for keyword in self.keywords:
                if text[last:].startswith(keyword):
                    return text[pos:last]
            last += 1

    def __init__(self, language: ProofAction.Language = ProofAction.Language.COQ):
        self.language = language
        if language == ProofAction.Language.COQ:
            self.keywords = [FewShotGptCoqKeywords.THEOREM, FewShotGptCoqKeywords.DEFINITION, FewShotGptCoqKeywords.DEFINITIONS, FewShotGptCoqKeywords.LEMMA, FewShotGptCoqKeywords.LEMMAS, FewShotGptCoqKeywords.END]
        elif language == ProofAction.Language.LEAN:
            self.keywords = [FewShotGptLeanKeywords.THEOREM, FewShotGptLeanKeywords.DEFINITION, FewShotGptLeanKeywords.DEFINITIONS, FewShotGptLeanKeywords.LEMMA, FewShotGptLeanKeywords.LEMMAS, FewShotGptLeanKeywords.END, FewShotGptLeanKeywords.INFORMAL_THEOREM, FewShotGptLeanKeywords.INFORMAL_PROOF]
        else:
            raise NotImplementedError(f"language {language} not supported")
        recognizers = {
            'String': self.before_keyword
        }
        if language == ProofAction.Language.COQ:
            terminals = f"""
terminals
End: "{FewShotGptCoqKeywords.END}";
Thm: "{FewShotGptCoqKeywords.THEOREM}";
Dfn: "{FewShotGptCoqKeywords.DEFINITION}";
Dfns: "{FewShotGptCoqKeywords.DEFINITIONS}";
Lm: "{FewShotGptCoqKeywords.LEMMA}";
Lms: "{FewShotGptCoqKeywords.LEMMAS}";
String:;
"""
        elif language == ProofAction.Language.LEAN:
            terminals = f"""
terminals
End: "{FewShotGptLeanKeywords.END}";
Thm: "{FewShotGptLeanKeywords.THEOREM}";
Dfn: "{FewShotGptLeanKeywords.DEFINITION}";
Dfns: "{FewShotGptLeanKeywords.DEFINITIONS}";
Lm: "{FewShotGptLeanKeywords.LEMMA}";
Lms: "{FewShotGptLeanKeywords.LEMMAS}"
InfThm: "{FewShotGptLeanKeywords.INFORMAL_THEOREM}"
InfPrf: "{FewShotGptLeanKeywords.INFORMAL_PROOF}";
String:;
"""
        else:
            raise NotImplementedError(f"language {language} not supported")
        if language == ProofAction.Language.COQ:
            self.END = FewShotGptCoqKeywords.END
            self.THEOREM = FewShotGptCoqKeywords.THEOREM
            self.DEFINITION = FewShotGptCoqKeywords.DEFINITION
            self.DEFINITIONS = FewShotGptCoqKeywords.DEFINITIONS
            self.LEMMA = FewShotGptCoqKeywords.LEMMA
            self.LEMMAS = FewShotGptCoqKeywords.LEMMAS
        elif language == ProofAction.Language.LEAN:
            self.END = FewShotGptLeanKeywords.END
            self.THEOREM = FewShotGptLeanKeywords.THEOREM
            self.DEFINITION = FewShotGptLeanKeywords.DEFINITION
            self.DEFINITIONS = FewShotGptLeanKeywords.DEFINITIONS
            self.LEMMA = FewShotGptLeanKeywords.LEMMA
            self.LEMMAS = FewShotGptLeanKeywords.LEMMAS
            self.INFORMAL_THEOREM = FewShotGptLeanKeywords.INFORMAL_THEOREM
            self.INFORMAL_PROOF = FewShotGptLeanKeywords.INFORMAL_PROOF
        else:
            raise NotImplementedError(f"language {language} not supported")
        grammar = FewShotGptResponseGrammar.grammar + terminals
        super(FewShotGptResponseGrammar, self).__init__(grammar, self.keywords, recognizers=recognizers)
    
    def format_as_per_grammar(self, coq_gpt_response: FewShotGptResponse, k: typing.Optional[int] = None, max_token_cnt: typing.Optional[int] = None, characters_per_token: int = 4) -> str:
        assert coq_gpt_response.theorem is not None
        char_cnt = max_token_cnt * characters_per_token if max_token_cnt is not None else None # 4 is the average length of a token as per OpenAI
        text = ""
        lines_map = {
            self.THEOREM : [],
            self.DEFINITIONS : [],
            self.LEMMAS : [],
        }
        lines_order = [
            self.THEOREM,
            self.DEFINITIONS,
            self.LEMMAS
        ]
        priority_order_lo_hi = [
            self.LEMMAS,
            self.DEFINITIONS,
            self.THEOREM
        ]
        if self.language == ProofAction.Language.LEAN:
            lines_map[self.INFORMAL_THEOREM] = []
            lines_map[self.INFORMAL_PROOF] = []
            lines_order = [
                self.THEOREM,
                self.INFORMAL_THEOREM,
                self.INFORMAL_PROOF,
                self.DEFINITIONS,
                self.LEMMAS
            ]
            priority_order_lo_hi = [
                self.LEMMAS,
                self.DEFINITIONS,
                self.INFORMAL_PROOF,
                self.INFORMAL_THEOREM,
                self.THEOREM
            ]
        new_line = f"{self.THEOREM}\n{coq_gpt_response.theorem}"
        lines_map[self.THEOREM] = [new_line]

        if len(coq_gpt_response.defintions) > 0:
            lines_map[self.DEFINITIONS] = [f"\n{self.DEFINITIONS}"]
        for idx, dfn in enumerate(coq_gpt_response.defintions):
            if k is not None and idx >= k:
                break
            lines_map[self.DEFINITIONS].append(f"{self.DEFINITION} {dfn}")

        if len(coq_gpt_response.lemmas) > 0:
            lines_map[self.LEMMAS] = [f"\n{self.LEMMAS}"]
        for idx, lm in enumerate(coq_gpt_response.lemmas):
            if k is not None and idx >= k:
                break
            lines_map[self.LEMMAS].append(f"{self.LEMMA} {lm}")
        
        if self.language == ProofAction.Language.LEAN:
            if coq_gpt_response.informal_theorem is not None:
                lines_map[self.INFORMAL_THEOREM] = ["\n" + self.INFORMAL_THEOREM]
                lines_map[self.INFORMAL_THEOREM].append(coq_gpt_response.informal_theorem)
            if coq_gpt_response.informal_proof is not None:
                lines_map[self.INFORMAL_PROOF] = [self.INFORMAL_PROOF]
                lines_map[self.INFORMAL_PROOF].append(coq_gpt_response.informal_proof)

        keywords = [keyword for keyword in lines_map.keys()]
        # Convert all the lines under each keyword to a single string
        for keyword in keywords:
            lines_map[keyword] = "\n".join(lines_map[keyword])
        # Frame the first prompt version without any token limit
        text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0])
        
        # Trim the lines based on the max_token_cnt
        if char_cnt is not None and len(text) > char_cnt:
            _idx = 0
            diff = len(text) - char_cnt
            while _idx < len(priority_order_lo_hi) and diff > 0:
                trim_part = priority_order_lo_hi[_idx]
                if trim_part in lines_map:
                    lines_map[trim_part] = lines_map[trim_part][:-diff] # Trim everything except the STEPS from the end
                text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0])
                diff = len(text) - char_cnt
                _idx += 1
        text += f"\n\n{self.END}"
        # verify that the text is valid as per grammar by compiling it
        # self.compile(text)
        if char_cnt is not None and len(text) > char_cnt:
            text = text[:char_cnt] # Just trim the text from the end because no trimming strategy has worked out
        return text


class FewShotGptRequestGrammar(Grammar):
    grammar = f"""
Prog: 
    Proof String Qed;
"""

    def before_keyword(self, text, pos):
        last = pos
        while last < len(text):
            for keyword in self.keywords:
                if text[last:].startswith(keyword):
                    return text[pos:last]
            last += 1

    def __init__(self, language: ProofAction.Language = ProofAction.Language.COQ, enable_defensive_parsing: bool = False):
        self.language = language
        if language == ProofAction.Language.COQ:
            self.keywords = [FewShotGptCoqKeywords.PROOF, FewShotGptCoqKeywords.QED]
        elif language == ProofAction.Language.LEAN:
            self.keywords = [FewShotGptLeanKeywords.PROOF, FewShotGptLeanKeywords.QED]
        else:
            raise NotImplementedError(f"language {language} not supported")
        recognizers = {
            'String': self.before_keyword
        }
        if self.language == ProofAction.Language.COQ:
            terminals = f"""
terminals
Proof: "{FewShotGptCoqKeywords.PROOF}";
Qed: "{FewShotGptCoqKeywords.QED}";
String:;
"""
        elif self.language == ProofAction.Language.LEAN:
            terminals = f"""
terminals
Proof: "{FewShotGptLeanKeywords.PROOF}";
Qed: "{FewShotGptLeanKeywords.QED}";
String:;
"""
        else:
            raise NotImplementedError(f"language {language} not supported")
        if language == ProofAction.Language.COQ:
            self.PROOF = FewShotGptCoqKeywords.PROOF
            self.QED = FewShotGptCoqKeywords.QED
        elif language == ProofAction.Language.LEAN:
            self.PROOF = FewShotGptLeanKeywords.PROOF
            self.QED = FewShotGptLeanKeywords.QED
        else:
            raise NotImplementedError(f"language {language} not supported")
        grammar = FewShotGptRequestGrammar.grammar + terminals
        self.enable_defensive_parsing = enable_defensive_parsing
        super(FewShotGptRequestGrammar, self).__init__(grammar, self.keywords, recognizers=recognizers)

    def _parse_expr(self, nonTerminal: str, nodes) -> FewShotGptRequest:
        if nonTerminal == "Prog":
            assert len(nodes) >= 3
            if self.language == ProofAction.Language.COQ:
                actions = str(nodes[1]).strip() + f"\n{self.QED}"
            elif self.language == ProofAction.Language.LEAN:
                actions = str(nodes[1]).strip()
                if actions.startswith("begin"):
                    actions = actions[len("begin"):]
                if not actions.endswith("end"):
                    actions += "end"
            else:
                raise NotImplementedError(f"language {self.language} not supported")
            proof_action = ProofAction(ProofAction.ActionType.RUN_TACTIC, self.language, tactics=[actions])
            return FewShotGptRequest(action=proof_action, proof_string=actions)
        else:
            raise Exception(f"Unknown non-terminal {nonTerminal}")
        
    def get_action(self, inp=None):
        actions = {
            "Prog": lambda _, nodes: self._parse_expr('Prog', nodes)
        }
        return actions
    
    def interpret_result(self, result):
        assert isinstance(result, FewShotGptRequest), f"Result must be a FewShotGptRequest. Got {type(result)}"
        return result
    
    def generate_message_from_gpt_request(self, coq_gpt_request: FewShotGptRequest) -> str:
        return f"{self.PROOF}\n{coq_gpt_request.proof_string}"

    def get_openai_request(self, message_response: str) -> typing.Tuple[FewShotGptRequest, str]:
        message, _ = message_response
        if self.enable_defensive_parsing:
            return self.defensive_parsing(message)
        else:
            return self.normal_parsing(message)
        
    def normal_parsing(self, message):
        message_temp = message
        message_temp += f"\n{self.QED}"
        result : FewShotGptRequest = self.run(message_temp, None)            
        message_temp = self.generate_message_from_gpt_request(result)
        return (result, message_temp)

    def defensive_parsing(self, message):
        start_idx = 0
        end_idx = len(message)
        # Generate all possible sub-strings such that the start_idx is less than end_idx
        idxs = [(s_idx, e_idx) for s_idx in range(start_idx, end_idx) for e_idx in range(end_idx, s_idx, -1)]
        message_temp = message
        message_parsed = False
        for s_idx, e_idx in idxs:
            # This type of robust parsing can be needed in case of some LLMs which
            # don't follow the specified format
            try:
                message_temp = message[s_idx:e_idx]
                if message_temp.endswith(self.QED):
                    # Just in case the LLM doesn't remove the stop token
                    message_temp = message_temp.strip(self.QED)
                message_temp += f"\n{self.QED}"
                result : FewShotGptRequest = self.run(message_temp, None)            
                message_temp = self.generate_message_from_gpt_request(result)
                message_parsed = True
            except:
                message_parsed = False
            if message_parsed:
                break
        if not message_parsed:
            message_temp = message[start_idx:end_idx]
            message_temp += f"\n{self.QED}"
            result : FewShotGptRequest = self.run(message_temp, None)            
            message_temp = self.generate_message_from_gpt_request(result)
        return (result, message_temp)
    
    def parse_request_to_args(self, messages: typing.List[str]) -> typing.List[str]:
        results : typing.List[str] = []
        for message in messages:
            assert message.endswith(self.QED), "Message must end with end token"
            result : FewShotGptRequest = self.run(message, None)
            results.extend(result.args)
        return results

class InformalFewShotGptLeanKeywords(object):
    THEOREM = "[THEOREM]"
    PROOF = "[PROOF]"
    END = "[END]"

@dataclass_json
@dataclass
class InformalFewShotGptResponse(object):
    theorem: str

class InformalFewShotGptResponseGrammar(Grammar):
    grammar = f"""
Prog:
  Theorem String End;
"""

    def before_keyword(self, text, pos):
        last = pos
        while last < len(text):
            for keyword in self.keywords:
                if text[last:].startswith(keyword):
                    return text[pos:last]
            last += 1

    def __init__(self, language: ProofAction.Language = ProofAction.Language.LEAN):
        self.language = language
        if language == ProofAction.Language.LEAN:
            self.keywords = [InformalFewShotGptLeanKeywords.THEOREM, InformalFewShotGptLeanKeywords.END]
        else:
            raise NotImplementedError(f"language {language} not supported")
        recognizers = {
            'String': self.before_keyword
        }
        if language == ProofAction.Language.LEAN:
            terminals = f"""
terminals
End: "{InformalFewShotGptLeanKeywords.END}";
Theorem: "{InformalFewShotGptLeanKeywords.THEOREM}";
String:;
"""
        else:
            raise NotImplementedError(f"language {language} not supported")
        if language == ProofAction.Language.LEAN:
            self.END = InformalFewShotGptLeanKeywords.END
            self.THEOREM = InformalFewShotGptLeanKeywords.THEOREM
        else:
            raise NotImplementedError(f"language {language} not supported")
        grammar = InformalFewShotGptResponseGrammar.grammar + terminals
        super(InformalFewShotGptResponseGrammar, self).__init__(grammar, self.keywords, recognizers=recognizers)
    
    def format_as_per_grammar(self, coq_gpt_response: InformalFewShotGptResponse, k: typing.Optional[int] = None, max_token_cnt: typing.Optional[int] = None, characters_per_token: int = 4) -> str:
        assert coq_gpt_response.theorem is not None
        char_cnt = max_token_cnt * characters_per_token if max_token_cnt is not None else None # 4 is the average length of a token as per OpenAI
        text = ""
        lines_map = {
            self.THEOREM : [],
            self.END : []
        }
        lines_order = [
            self.THEOREM,
            self.END
        ]
        priority_order_lo_hi = [
            self.THEOREM,
            self.END
        ]
        
        new_line = f"{self.THEOREM}\n{coq_gpt_response.theorem}"
        lines_map[self.THEOREM] = [new_line]

        keywords = [keyword for keyword in lines_map.keys()]
        # Convert all the lines under each keyword to a single string
        for keyword in keywords:
            lines_map[keyword] = "\n".join(lines_map[keyword])
        # Frame the first prompt version without any token limit
        text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0])
        
        # Trim the lines based on the max_token_cnt
        if char_cnt is not None and len(text) > char_cnt:
            _idx = 0
            diff = len(text) - char_cnt
            while _idx < len(priority_order_lo_hi) and diff > 0:
                trim_part = priority_order_lo_hi[_idx]
                if trim_part in lines_map:
                    lines_map[trim_part] = lines_map[trim_part][:-diff] # Trim everything except the STEPS from the end
                text = "\n".join([lines_map[keyword] for keyword in lines_order if keyword in lines_map if len(lines_map[keyword]) > 0])
                diff = len(text) - char_cnt
                _idx += 1
        text += f"\n\n{self.END}"
        # verify that the text is valid as per grammar by compiling it
        # self.compile(text)
        if char_cnt is not None and len(text) > char_cnt:
            text = text[:char_cnt] # Just trim the text from the end because no trimming strategy has worked out
        return text


class InformalFewShotGptRequestGrammar(Grammar):
    grammar = f"""
Prog: 
    Proof String End;
"""

    def before_keyword(self, text, pos):
        last = pos
        while last < len(text):
            for keyword in self.keywords:
                if text[last:].startswith(keyword):
                    return text[pos:last]
            last += 1

    def __init__(self, language: ProofAction.Language = ProofAction.Language.LEAN, enable_defensive_parsing: bool = False):
        self.language = language
        if language == ProofAction.Language.LEAN:
            self.keywords = [InformalFewShotGptLeanKeywords.PROOF, InformalFewShotGptLeanKeywords.END]
        else:
            raise NotImplementedError(f"language {language} not supported")
        recognizers = {
            'String': self.before_keyword
        }
        if self.language == ProofAction.Language.LEAN:
            terminals = f"""
terminals
Proof: "{InformalFewShotGptLeanKeywords.PROOF}";
End: "{InformalFewShotGptLeanKeywords.END}";
String:;
"""
        else:
            raise NotImplementedError(f"language {language} not supported")
        if language == ProofAction.Language.LEAN:
            self.PROOF = InformalFewShotGptLeanKeywords.PROOF
            self.END = InformalFewShotGptLeanKeywords.END
        else:
            raise NotImplementedError(f"language {language} not supported")
        grammar = InformalFewShotGptRequestGrammar.grammar + terminals
        self.enable_defensive_parsing = enable_defensive_parsing
        super(InformalFewShotGptRequestGrammar, self).__init__(grammar, self.keywords, recognizers=recognizers)

    def _parse_expr(self, nonTerminal: str, nodes) -> ProofAction:
        if nonTerminal == "Prog":
            assert len(nodes) >= 3
            proof = str(nodes[1]).strip()
            return ProofAction(action_type=ProofAction.ActionType.INFORMAL, language=self.language, proof=proof)
        else:
            raise Exception(f"Unknown non-terminal {nonTerminal}")
        
    def get_action(self, inp=None):
        actions = {
            "Prog": lambda _, nodes: self._parse_expr('Prog', nodes)
        }
        return actions
    
    def interpret_result(self, result):
        assert isinstance(result, ProofAction), f"Result must be a ProofAction. Got {type(result)}"
        return result
    
    def generate_message_from_gpt_request(self, coq_gpt_request: ProofAction) -> str:
        assert coq_gpt_request.action_type == ProofAction.ActionType.INFORMAL, f"action_type must be {ProofAction.ActionType.INFORMAL}, not {coq_gpt_request.action_type}"
        return f"{self.PROOF}\n{coq_gpt_request.kwargs['proof']}\n{self.END}"

    def get_openai_request(self, message_response: str) -> typing.Tuple[ProofAction, str]:
        message, _ = message_response
        if self.enable_defensive_parsing:
            return self.defensive_parsing(message)
        else:
            return self.normal_parsing(message)
    
    def normal_parsing(self, message):
        message_temp = message
        message_temp += f"\n{self.END}"
        result : ProofAction = self.run(message_temp, None)            
        message_temp = self.generate_message_from_gpt_request(result)
        return (result, message_temp)

    def defensive_parsing(self, message):
        start_idx = 0
        end_idx = len(message)
        # Generate all possible sub-strings such that the start_idx is less than end_idx
        idxs = [(s_idx, e_idx) for s_idx in range(start_idx, end_idx) for e_idx in range(end_idx, s_idx, -1)]
        message_temp = message
        message_parsed = False
        for s_idx, e_idx in idxs:
            # This type of robust parsing can be needed in case of some LLMs which
            # don't follow the specified format
            try:
                message_temp = message[s_idx:e_idx]
                if message_temp.endswith(self.END):
                    # Just in case the LLM doesn't remove the stop token
                    message_temp = message_temp.strip(self.END)
                message_temp += f"\n{self.END}"
                result : ProofAction = self.run(message_temp, None)            
                message_temp = self.generate_message_from_gpt_request(result)
                message_parsed = True
            except:
                message_parsed = False
            if message_parsed:
                break
        if not message_parsed:
            message_temp = message[start_idx:end_idx]
            message_temp += f"\n{self.END}"
            result : ProofAction = self.run(message_temp, None)            
            message_temp = self.generate_message_from_gpt_request(result)
        return (result, message_temp)
    
    def parse_request_to_args(self, messages: typing.List[str]) -> typing.List[str]:
        results : typing.List[str] = []
        for message in messages:
            assert message.endswith(self.END), "Message must end with end token"
            result : ProofAction = self.run(message, None)
            results.extend(result.args)
        return results


#!/usr/bin/env python3
import os
import typing
try:
    from .grammar import Grammar
except ImportError:
    from grammar import Grammar

class PromptGrammar(Grammar):
    grammar = """
Prog:
  ConvStart Convs1 ConvEnd
| ConvStart Convs2 ConvEnd;
Convs1:
  Conv1
| Conv1 Conv2
| Conv1 Conv2 Convs1;
Convs2:
  Conv2
| Conv2 Conv1
| Conv2 Conv1 Convs2;
Conv1:
    '`{0}`' String;
Conv2:
    '`{1}`' String;

terminals
ConvStart: "`conv start`";
ConvEnd: "`conv end`";
String:;
"""

    def get_string_parser(keywords):
        def string_parser(keywords: typing.List[str], text: str, pos: int):
            last = pos
            while last < len(text):
                while last < len(text) and text[last] != '`':
                    last += 1
                if last < len(text):
                    for keyword in keywords:
                        if text[last:].startswith(keyword):
                            return text[pos:last]
                    last += 1
        return lambda text, pos: string_parser(keywords, text, pos)

    def __init__(self, user_name: str = 'example_user', agent_name: str = 'example_assistant'):
        self.keywords = [f"`{user_name}`", f"`{agent_name}`", "`conv start`", "`conv end`"]
        self.user_name = user_name
        self.agent_name = agent_name
        recognizers = {
            'String': PromptGrammar.get_string_parser(self.keywords)
        }
        super(PromptGrammar, self).__init__(PromptGrammar.grammar.format(user_name, agent_name), self.keywords, recognizers=recognizers)
    pass

    def _parse_expr(self, nonTerminal, nodes, role, context: typing.List[dict]):
        if nonTerminal == "Conv1" or nonTerminal == "Conv2":
            assert len(nodes) == 2
            context.append({
                'role': role,
                'name': str(nodes[0]).strip('`'),
                'content' : nodes[1].strip()
            })
        else:
            raise Exception(f"Unknown non-terminal {nonTerminal}")
    
    def get_action(self, inp):
        assert isinstance(inp, str), f"Input must be a string. Got {type(inp)}"
        assert inp in {"system", "user"}, f"Input must be either 'system' or 'user'. Got {inp}"
        role = inp
        context = []
        actions = {
            "Prog": lambda _, nodes: context,
            "Conv1": lambda _, nodes: self._parse_expr('Conv1', nodes, role, context),
            "Conv2": lambda _, nodes: self._parse_expr('Conv2', nodes, role, context),
            "String": lambda _, nodes: str(nodes) # Since this is always a string
        }
        return actions
    
    def interpret_result(self, result):
        assert isinstance(result, list), f"Result must be a list. Got {type(result)}"
        return result
    
    def get_conv_messages(self, file_path: str, role: str = "system"):
        assert os.path.exists(file_path), f"File {file_path} does not exist"
        with open(file_path, "r") as f:
            conv = f.read()
        result = self.run(conv, role)
        return result

    def get_main_message(self, file_path: str, role: str = "system"):
        with open(file_path) as f:
            main_content = f.read()
        return self.get_main_message_from_string(main_content, role)

    def get_main_message_from_string(self, content: str, role: str = "system"):
        return {
            'role': role,
            'content': content
        }

    def parse_messages(self, messages: list, role: str = "assistant"):
        assert isinstance(messages, list), f"Messages must be a list. Got {type(messages)}"
        result = [(message["content"], message["finish_reason"]) for message in messages if message['role'] == role]
        return result




#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
import os
from proof_wala.llm_helpers.model import TrainingDataFormatterCallback
from proof_wala.llm_helpers.llama2_chat_format import Llama2ChatFormat
from proof_wala.llm_helpers.theorem_proving_training_dataset import TheoremProvingTrainingDataset
from itp_interface.tools.training_data import TrainingData
from proof_wala.parsers.copra_grammars import CoqGPTResponseDfsGrammar, CoqGPTRequestGrammar, CoqGptRequest, CoqGptResponse
from proof_wala.parsers.grammars.prompt_template_grammar import PromptGrammar

class CopraTrainingDataset(TheoremProvingTrainingDataset):
    def __init__(self, 
            training_data: TrainingData, 
            system_prompt_file: str, 
            conversation_prompt_file: str,
            characters_per_token: float = 3.6,
            max_tokens: int = 2048):
        assert os.path.exists(system_prompt_file), f"System prompt file {system_prompt_file} does not exist"
        assert os.path.exists(conversation_prompt_file), f"Conversation prompt file {conversation_prompt_file} does not exist"
        assert characters_per_token >= 1, f"Characters per token must be at least 1, but is {characters_per_token}"
        super().__init__(training_data)
        self.response_grammar = CoqGPTResponseDfsGrammar()
        self.request_grammar = CoqGPTRequestGrammar()
        self.system_prompt_file = system_prompt_file
        self.conversation_prompt_file = conversation_prompt_file
        self.prompt_grammar = PromptGrammar()
        self.system_message = self.prompt_grammar.get_main_message(self.system_prompt_file)
        self.characters_per_token = characters_per_token
        self.conversation_messages = self.prompt_grammar.get_conv_messages(self.conversation_prompt_file)
        self.system_message_chars_length = self._get_message_length(self.system_message)
        self.conversation_message_chars_length = self._get_message_length(self.conversation_messages)
        self.system_message_token_length = int(self.system_message_chars_length / self.characters_per_token)
        self.conversation_message_token_length = int(self.conversation_message_chars_length / self.characters_per_token)
        self.max_tokens = max_tokens
        self.max_tokens_in_prompt = self.max_tokens - self.system_message_token_length - self.conversation_message_token_length
        self.max_chars_in_prompt = int(self.max_tokens_in_prompt * self.characters_per_token)
        if self.max_tokens_in_prompt < 0:
            raise ValueError(f"Max tokens in prompt is negative: {self.max_tokens_in_prompt}, increase max tokens or decrease characters per token")
        self.llama2_chat_format = Llama2ChatFormat()

    def __getitem__(self, idx):
        response, request = self._get_response_request(idx)
        prompt = self.response_grammar.format_as_per_grammar(response, max_token_cnt=self.max_tokens_in_prompt, characters_per_token=self.characters_per_token)
        completion = self.request_grammar.generate_message_from_gpt_request(request)
        prompt_msg = self.prompt_grammar.get_main_message_from_string(prompt, role="user")
        completion_msg = self.prompt_grammar.get_main_message_from_string(completion, role="assistant")  
        messages = [self.system_message] + self.conversation_messages + [prompt_msg]
        partial_prompt, _ = self.llama2_chat_format(messages)
        completion_msg, _ = self.llama2_chat_format([completion_msg])
        return {
            "prompt": partial_prompt,
            "completion": completion_msg
        }
    
    def _get_message_length(self, message):
        # for base types like str, int, float, etc. the length is the number of characters in the string representation
        # for lists, the length is the sum of the lengths of the elements
        # for dicts, the length is the sum of the lengths of the keys and values
        # for tuples, the length is the sum of the lengths of the elements
        if isinstance(message, str):
            return len(message)
        elif isinstance(message, int):
            return len(str(message))
        elif isinstance(message, float):
            return len(str(message))
        elif isinstance(message, list):
            return sum([self._get_message_length(element) for element in message])
        elif isinstance(message, dict):
            return sum([self._get_message_length(key) + self._get_message_length(value) for key, value in message.items()])
        elif isinstance(message, tuple):
            return sum([self._get_message_length(element) for element in message])
        else:
            raise Exception(f"Unknown type {type(message)}")

    def _get_response_request(self, idx):
        last_tdf = self.training_data[idx]
        tdf = self.training_data[idx - 1] if idx > 0 else None
        last_step = tdf.proof_steps if tdf is not None and tdf.proof_id == last_tdf.proof_id else None
        tdf = self.training_data[idx - 2] if idx > 1 else None
        steps : typing.List[str] = []
        while tdf is not None and last_tdf.proof_id == tdf.proof_id:
            steps.extend(reversed(tdf.proof_steps))
            idx -= 1
            tdf = self.training_data[idx - 1] if idx > 1 else None
        steps.reverse()
        response = CoqGptResponse(
            success=True,
            steps=steps,
            last_step='\n'.join(last_step) if last_step is not None else None,
            training_data_format=last_tdf
        )
        request = CoqGptRequest(
            args=last_tdf.proof_steps
        )
        return response, request
    

class CopraPromptTrainingDataFormatter(TrainingDataFormatterCallback):
    def __call__(self, training_data_format_examples: typing.List[typing.Dict[str, str]]) -> typing.List[str]:
        output_texts = []
        all_prompts = training_data_format_examples["prompt"]
        all_completions = training_data_format_examples["completion"]
        stopping_token = "" # The stopping token is already included in the completion because LLama2ChatFormat adds it
        for prompt, completion in zip(all_prompts, all_completions):
            output_text = f"{prompt}{completion}{stopping_token}"
            output_texts.append(output_text)
        return output_texts
    
    def get_prompt_and_completion(self, training_data_examples: typing.List[typing.Dict[str, str]]) -> typing.List[typing.Tuple[str, str]]:
        all_prompts = training_data_examples["prompt"]
        all_completions = training_data_examples["completion"]
        prompt_and_completions = []
        if isinstance(all_prompts, str):
            all_prompts = [all_prompts]
            assert isinstance(all_completions, str)
            all_completions = [all_completions]
        assert len(all_prompts) == len(all_completions)
        for prompt, completion in zip(all_prompts, all_completions):
            # Trim everything after the stopping token from the completion
            completion = completion.split(self.get_stopping_token())[0]
            prompt_and_completion = (prompt, completion + self.get_stopping_token())
            prompt_and_completions.append(prompt_and_completion)
        return prompt_and_completions
    
    def get_stopping_token(self):
        return "[END]"

if __name__ == "__main__":
    prompt_name = "copra-coq-dfs"
    system_prompt_file = f"src/proof_wala/data/prompts/system/{prompt_name}.md"
    conversation_prompt_file = f"src/proof_wala/data/prompts/conversation/{prompt_name}.md"
    data_folder = f".log/train"
    meta_filename = "local.meta.json"
    training_data = TrainingData(data_folder, meta_filename)
    with CopraTrainingDataset(training_data, system_prompt_file, conversation_prompt_file, max_tokens=4096) as dataset:
        hf_dataset = dataset.get_hf_dataset()
        formatter = CopraPromptTrainingDataFormatter()
        formatted_dataset = formatter(hf_dataset)
        for example in formatted_dataset:
            print(example)
        prompt_and_completions = formatter.get_prompt_and_completion(hf_dataset)
    pass


import unittest
import torch

class TestProofWalaCI(unittest.TestCase):
    def test_torch_installation(self):
        """Test that torch is installed and has a version, and print CUDA availability."""
        version = torch.__version__
        self.assertIsNotNone(version, "Torch version should not be None.")
        print("Torch version:", version)
        print("CUDA available:", torch.cuda.is_available())
    
    def test_proof_wala_import(self):
        """Test that the proof_wala package can be imported."""
        try:
            import proof_wala
        except Exception as e:
            self.fail("Failed to import proof_wala package: " + str(e))
    
    def test_itp_interface_import(self):
        """Optionally test that the itp-interface dependency is available.
        
        If itp-interface is not installed, skip the test.
        """
        try:
            import itp_interface  # or the actual module name if different
        except ImportError as e:
            self.skipTest("itp-interface not installed: " + str(e))
    
    def test_huggingface_pipeline(self):
        """Test that a transformers pipeline for the ProofWala model can be instantiated.
        
        We use local_files_only=True to avoid network access during CI.
        If the model files are not cached locally, the test is skipped.
        """
        try:
            from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, pipeline
            model_name = "amitayusht/ProofWala-Multilingual"
            # Use local_files_only=True to ensure CI does not require network access.
            tokenizer = AutoTokenizer.from_pretrained(model_name, local_files_only=True)
            model = AutoModelForSeq2SeqLM.from_pretrained(model_name, local_files_only=True)
            text2text_pipe = pipeline("text2text-generation", model=model, tokenizer=tokenizer, device=-1)
            test_input = """Goals to prove:
[GOALS]
[GOAL] 1
forall n : nat, n + 1 = 1 + n
[END]"""
            output = text2text_pipe(test_input, max_length=100)
            print("Output:", output)
            self.assertIsInstance(output, list, "Expected output to be a list.")
        except Exception as e:
            self.skipTest("Skipping HuggingFace pipeline test: " + str(e))

if __name__ == '__main__':
    unittest.main()



import time
from multiprocessing import Pool
import os
import typing
import logging
try:
    from .search import Node, Edge, SearchAlgorithm
except ImportError:
    from search import Node, Edge, SearchAlgorithm

class BeamSearch(SearchAlgorithm):
    def __init__(self, beam_width: int = 1):
        assert beam_width > 0, "Beam width must be greater than 0"
        super().__init__()
        self.beam_width = beam_width  # The maximum number of nodes to consider at each level

    def search(
            self,
            start: Node, 
            goal: Node,
            heuristic: typing.Callable[[Node, Edge, Node], float], 
            generate_children: typing.Callable[[Node, float], typing.Tuple[typing.List[Node], typing.List[Edge]]] = None,
            parallel_count: int = None,
            build_tree: bool = True,
            timeout_in_secs: float = None,
            logger: logging.Logger = None) -> typing.Tuple[Node, bool, float]:

        assert (generate_children is not None and build_tree) or not build_tree, "Must provide generate_children function"
        time_elapsed = 0
        start_time = time.time()
        timeout_in_secs = timeout_in_secs if timeout_in_secs else float('inf')
        logger = logger if logger else logging.getLogger(__name__)

        # Instead of a min heap, use a list to maintain the current level's nodes
        current_level = [(heuristic(None, None, start), start)]
        next_level = []
        explored = set([start])  # Keep track of explored nodes to avoid cycles

        # Initialize multiprocessing pool
        pool = Pool(processes=parallel_count if parallel_count else max(os.cpu_count() - 1, 1))
        start.cummulative_score = 0
        start.distance_from_root = 0
        tree_nodes : typing.Dict[Node, Node] = {}  # Map of all nodes in the tree
        tree_nodes[start] = start
        should_stop = False
        while current_level:
            time_elapsed = time.time() - start_time
            if time_elapsed > timeout_in_secs:
                pool.close()
                pool.join()
                return (start, False, time_elapsed)
            new_tree_nodes : typing.List[typing.Tuple[Node, Node, Edge]] = []

            # For early stopping, check if the goal node is in the current level
            logger.info(f"Frontier size: {len(current_level)}")
            # logger.info("Dumping the frontier nodes")
            # logger.info("-" * 50)
            for _, current_node in current_level:
                # logger.info(f"Frontier node distance from root: {current_node.distance_from_root}, frontier node:\n {current_node}")
                if current_node == goal:
                    pool.close()
                    pool.join()
                    time_elapsed = time.time() - start_time
                    assert len(current_node.parents) > 0, f"Goal node must have at least one parent"
                    return (current_node, True, time_elapsed)
            # logger.info("Dumped the frontier")
            # logger.info("-" * 50)

            for _, current_node in current_level:
                if build_tree:
                    remaining_timeout = timeout_in_secs - time_elapsed
                    children, edges = generate_children(current_node, remaining_timeout)
                    time_elapsed = time.time() - start_time
                else:
                    children = current_node.children
                    edges = current_node.edges
                children_to_explore : typing.Set[typing.Tuple[Node, Edge]] = set()
                for child, edge in zip(children, edges):
                    if child not in explored:
                        children_to_explore.add((child, edge))
                    if build_tree:
                        new_tree_nodes.append((current_node, child, edge))

                unique_children_edge = list(children_to_explore)
                child_costs = pool.starmap(heuristic, [[current_node, edge, child] for child, edge in unique_children_edge])
                for (child, _), cost in zip(unique_children_edge, child_costs):
                    if child not in tree_nodes:
                        next_level.append((cost, child))
                # Iterate over the children to check if we found the goal node for early stopping
                for child, edge in zip(children, edges):
                    if child == goal:
                        should_stop = True
                if should_stop:
                    break
            
            current_level = next_level
            next_level = []
            current_level.sort(key=lambda x: x[0])  # Sort by heuristic value
            current_level = current_level[:self.beam_width]  # Keep only the top beam_width nodes
            beam_nodes = set([n for _, n in current_level])
            nodes_part_of_tree = []
            if build_tree:
                # Go over the tree_map and add the children to the current node
                for parent, child, edge in new_tree_nodes:
                    parent_distance = parent.distance_from_root
                    parent_cummulative = parent.cummulative_score
                    if child in beam_nodes:
                        if child not in tree_nodes:
                            child.distance_from_root = parent_distance + 1
                            child.cummulative_score = parent_cummulative + edge.score
                            parent.add_child(child, edge)
                            tree_nodes[child] = child
                            explored.add(child)
                            nodes_part_of_tree.append(child)
                        else:
                            child_idx = None
                            try:
                                child_idx = parent.children.index(child)
                            except ValueError:
                                pass
                            if child_idx is None:
                                existing_child = tree_nodes[child]
                                existing_child.distance_from_root = min(existing_child.distance_from_root, parent_distance + 1)
                                existing_child.score = min(existing_child.score, child.score)
                                existing_child.cummulative_score = min(existing_child.cummulative_score, parent_cummulative + edge.score)
                                parent.add_child(existing_child, edge)
                                nodes_part_of_tree.append(existing_child)
                            elif edge != parent.edges[child_idx]:
                                if edge not in parent.edges[child_idx].equivalent_edges:
                                    parent.edges[child_idx].add_equivalent_edge(edge)
                                    existing_child = parent.children[child_idx]
                                    existing_child.distance_from_root = min(existing_child.distance_from_root, parent_distance + 1)
                                    existing_child.score = min(existing_child.score, child.score)
                                    existing_child.cummulative_score = min(existing_child.cummulative_score, parent_cummulative + edge.score)
                current_level = [(node.cummulative_score, node) for node in nodes_part_of_tree]
                current_level.sort(key=lambda x: x[0])
    
        pool.close()
        pool.join()
        return (start, False, time_elapsed)



from transformers import (
    T5ForConditionalGeneration,
    T5Config,
    T5Tokenizer,
)

def get_empty_config(config_class_name):
    if config_class_name == "T5Config":
        return T5Config()
    else:
        raise ValueError(f"Unknown config class name: {config_class_name}")

def get_model_class(config_class_name):
    if config_class_name == "T5Config":
        return T5ForConditionalGeneration
    else:
        raise ValueError(f"Unknown model class name: {config_class_name}")

def trim_model_kwargs(model_kwargs, config_class_name):
    if config_class_name == "T5Config":
        return {k: v for k, v in model_kwargs.items() if k in T5ForConditionalGeneration.__init__.__code__.co_varnames}
    else:
        raise ValueError(f"Unknown model class name: {config_class_name}")

def get_decoder_start_token_id(config_class_name, tokenizer: T5Tokenizer):
    if config_class_name == "T5Config":
        return tokenizer.pad_token_id
    else:
        raise ValueError(f"Unknown model class: {config_class_name}")


#!/usr/bin/env python3

import torch
import numpy as np

class CudaContext:
    __cuda_context = None
    
    @staticmethod
    def get_default_context():
        if CudaContext.__cuda_context is None:
            CudaContext.__cuda_context = CudaContext()
        return CudaContext.__cuda_context
    
    @staticmethod
    def set_default_context(turn_off_cuda=False):
        CudaContext.__cuda_context = CudaContext(turn_off_cuda=turn_off_cuda)

    def __init__(self, turn_off_cuda=False):
        self.turn_off_cuda = turn_off_cuda

    def is_cuda_available(self):
        return not self.turn_off_cuda and torch.cuda.is_available()

    def free_cuda_cache_if_possible(self):
        if self.is_cuda_available():
            torch.cuda.empty_cache()

    def try_get_gpu(self, model):
        self.free_cuda_cache_if_possible()
        if isinstance(model, np.ndarray):
            return torch.from_numpy(model).cuda() if self.is_cuda_available() else torch.from_numpy(model)
        elif isinstance(model, torch.Tensor) or isinstance(model, torch.nn.Module):
            return model.cuda() if self.is_cuda_available() else model
        else:
            return model

    def get_param_count(self, model, only_trainable=False):
        return sum(p.numel() for p in model.parameters() if (not only_trainable) or p.requires_grad)

    def is_cuda_model(self, model):
        return next(model.parameters()).is_cuda

    def get_gpu_usage(self):
        return torch.cuda.get_device_properties("cuda:0").total_memory

    def get_gpu_memory_map(self):
        import subprocess
        result = subprocess.check_output(
            [
                'nvidia-smi', '--query-gpu=memory.used',
                '--format=csv,nounits,noheader'
            ], encoding='utf-8')
        gpu_memory = [str(int(_)/2**10) + " GiB" for _ in result.strip().split('\n')]
        gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
        return gpu_memory_map


def main():
    # Start the ray cluster
    from filelock import FileLock
    import json
    import os
    import ray
    import logging
    import time
    import sys
    import argparse
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("--num_cpus", type=int, default=100)
    argument_parser.add_argument("--object_store_memory", type=int, default=150*2**30)
    argument_parser.add_argument("--memory", type=int, default=300*2**30)
    argument_parser.add_argument("--metrics_report_interval_ms", type=int, default=3*10**8)
    args = argument_parser.parse_args()
    root_dir = f"{os.path.abspath(__file__).split('proof_wala')[-2]}"
    if root_dir not in sys.path:
        sys.path.append(root_dir)
    os.environ["PYTHONPATH"] = f"{root_dir}:{os.environ.get('PYTHONPATH', '')}"
    os.makedirs(".log/locks", exist_ok=True)
    os.makedirs(".log/ray", exist_ok=True)
    ray_was_started = False
    pid = os.getpid()
    print("Initializing Ray")
    print("PID: ", pid)
    with FileLock(".log/locks/ray.lock"):
        if os.path.exists(".log/ray/session_latest"):
            with open(".log/ray/session_latest", "r") as f:
                ray_session = f.read()
                ray_session = json.loads(ray_session)
            ray_address = ray_session["address"]
            ray.init(address=ray_address)
            print("Ray was already started")
            print("Ray session: ", ray_session)
        else:
            ray_session = ray.init(
                num_cpus=args.num_cpus, 
                object_store_memory=args.object_store_memory, 
                _memory=args.memory, 
                logging_level=logging.CRITICAL, 
                ignore_reinit_error=False, 
                log_to_driver=False, 
                configure_logging=False,
                _system_config={"metrics_report_interval_ms": args.metrics_report_interval_ms})
            ray_session = dict(ray_session)
            ray_session["main_pid"] = pid
            print("Ray session: ", ray_session)
            with open(".log/ray/session_latest", "w") as f:
                f.write(json.dumps(ray_session))
            ray_was_started = True
            print("Ray was started")
            print("Ray session: ", ray_session)
    # Flush the stdout buffer
    sys.stdout.flush()
    while ray_was_started:
        # Keep the ray cluster alive till killed
        time.sleep(10000)

if __name__ == "__main__":
    main()


#!/usr/bin/env python3

import os
import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import json
import time
import typing
import ray
import logging
from torch.utils.data import Dataset, DataLoader
from itp_interface.tools.training_data import TrainingData
from itp_interface.tools.log_utils import setup_logger
from proof_wala.llm_helpers.model import Model, TrainingDataFormatterCallback
from proof_wala.main.config import Experiment, ExperimentType

@ray.remote
def tokenize_and_count(experiment: Experiment, prompts: typing.List[str], completions: typing.List[str]) -> int:
    token_count = 0
    model = Model(
        experiment.model_settings.name_or_path, 
        experiment.training_settings.training_args, 
        log_folder=None,
        **experiment.model_settings.model_args)
    tokenizer, tokenizer_args = model.get_tokenizer()
    for prompt, completion in zip(prompts, completions):
        full_text = f"{prompt}{completion}"
        tokenized_input = model.tokenize(tokenizer, [full_text], tokenizer_args)
        token_count += sum([len(x) for x in tokenized_input["input_ids"]])
    return token_count

def token_count(experiment: Experiment, dataset: Dataset, formatter_callback: TrainingDataFormatterCallback, logger: logging.Logger) -> int:
    root_dir = f"{__file__.split('proof_wala')[0]}"
    os.environ["PYTHONPATH"] = f"{root_dir}:{os.environ.get('PYTHONPATH', '')}"
    batch_size = experiment.training_settings.training_args.train_batch_size
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
    all_prompts = []
    all_completions = []
    tokenization_results = []
    total_tokens = 0
    for batch_idx, dataset in enumerate(dataloader):
        prompts_and_completions = formatter_callback.get_prompt_and_completion(dataset)
        batch_prompts = [x[0] for x in prompts_and_completions]
        batch_completions = [x[1] for x in prompts_and_completions]
        all_prompts.extend(batch_prompts)
        all_completions.extend(batch_completions)
        tokenization_results.append(tokenize_and_count.remote(experiment, batch_prompts, batch_completions))
        if len(tokenization_results) >= 30:
            tokenization_results : typing.List[int] = ray.get(tokenization_results)
            total_tokens += sum(tokenization_results)
            tokenization_results = []
            logger.info(f"Processed {batch_idx * batch_size + len(batch_prompts)} examples")
            logger.info(f"Total tokens so far: {total_tokens}")
    tokenization_results : typing.List[int] = ray.get(tokenization_results)
    total_tokens += sum(tokenization_results)
    logger.info(f"Processed {batch_idx * batch_size + len(batch_prompts)} examples")
    logger.info(f"Total tokens so far: {total_tokens}")
    return total_tokens

def count_tokens(experiment: Experiment):
    assert experiment.expertiment_type == ExperimentType.TokenCount, f"Experiment type must be {ExperimentType.Training}, but is {experiment.expertiment_type}"
    assert experiment.training_data_settings.training_data_dir is not None, f"Training data directory must be specified"
    time_now = time.strftime("%Y%m%d-%H%M%S")
    training_dataset_log_dir = os.path.join(experiment.training_data_settings.training_data_log_dir, time_now)
    os.makedirs(training_dataset_log_dir, exist_ok=True)
    eval_dataset_log_dir = os.path.join(experiment.training_data_settings.eval_data_log_dir, time_now) if experiment.training_data_settings.eval_data_log_dir is not None else None
    if eval_dataset_log_dir is not None:
        os.makedirs(eval_dataset_log_dir, exist_ok=True)
    test_dataset_log_dir = os.path.join(experiment.training_data_settings.test_data_log_dir, time_now) if experiment.training_data_settings.test_data_log_dir is not None else None
    if test_dataset_log_dir is not None:
        os.makedirs(test_dataset_log_dir, exist_ok=True)
    should_load_eval = experiment.training_data_settings.eval_data_dir is not None
    should_load_test = experiment.training_data_settings.test_data_dir is not None
    should_split_train_eval = experiment.training_settings.train_eval_split
    if should_split_train_eval:
        assert not should_load_eval, f"Cannot split train and eval data if eval data is loaded from {experiment.training_data_settings.eval_data_dir}"
    # Load the training data
    training_data = TrainingData(experiment.training_data_settings.training_data_dir, experiment.training_data_settings.training_meta_filename, logger=setup_logger("TrainingData", os.path.join(training_dataset_log_dir, "training_data.log")))
    training_dataset_type = experiment.training_data_settings.training_dataset_type.get_class()
    with training_dataset_type(training_data, **experiment.training_data_settings.training_dataset_args) as training_dataset:
        training_data.logger.info(f"Loading training data from {experiment.training_data_settings.training_data_dir}")
        hf_training_dataset = training_dataset.get_hf_dataset()
        training_data.logger.info(f"Training data loaded with {len(hf_training_dataset)} examples")
        if should_split_train_eval:
            training_data.logger.info(f"Splitting training data into train and eval with eval percentage: {experiment.training_settings.eval_percentage}")
            hf_train_test_split = hf_training_dataset.train_test_split(test_size=experiment.training_settings.eval_percentage, seed=experiment.training_settings.training_args.data_seed)
            hf_training_dataset = hf_train_test_split["train"]
            hf_eval_dataset = hf_train_test_split["test"]
        else:
            hf_eval_dataset = None
        if experiment.training_settings.train_percentage < 1.0 and not should_split_train_eval:
            # Sample train_percentage of the dataset
            hf_training_dataset = hf_training_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(hf_training_dataset) * experiment.training_settings.train_percentage)))
    # Load the eval data
    if should_load_eval:
        eval_data = TrainingData(experiment.training_data_settings.eval_data_dir, experiment.training_data_settings.eval_meta_filename, logger=setup_logger("EvalData", os.path.join(eval_dataset_log_dir, "eval_data.log")))
        with training_dataset_type(eval_data, **experiment.training_data_settings.training_dataset_args) as eval_dataset:
            hf_eval_dataset = eval_dataset.get_hf_dataset()
    else:
        hf_eval_dataset = hf_eval_dataset if should_split_train_eval else None
    if hf_eval_dataset is not None:
        if experiment.training_settings.eval_percentage < 1.0 and not should_split_train_eval:
            # Sample eval_percentage of the dataset
            hf_eval_dataset = hf_eval_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(hf_eval_dataset) * experiment.training_settings.eval_percentage)))
    # Load the test data
    if should_load_test:
        test_data = TrainingData(experiment.training_data_settings.test_data_dir, experiment.training_data_settings.test_meta_filename, logger=setup_logger("TestData", os.path.join(test_dataset_log_dir, "test_data.log")))
        with training_dataset_type(test_data, **experiment.training_data_settings.training_dataset_args) as test_dataset:
            hf_test_dataset = test_dataset.get_hf_dataset()
    else:
        hf_test_dataset = None
    if hf_test_dataset is not None:
        if experiment.training_settings.test_percentage < 1.0:
            # Sample test_percentage of the dataset
            hf_test_dataset = hf_test_dataset.shuffle(seed=experiment.training_settings.training_args.data_seed).select(range(int(len(hf_test_dataset) * experiment.training_settings.test_percentage)))
    model_log_dir = os.path.join(experiment.model_settings.logging_dir, time_now)
    os.makedirs(model_log_dir, exist_ok=True)
    model_args = experiment.model_settings.model_args
    if "token" not in model_args or model_args["token"] is None:
        # try to get from .secrets/huggingface_token.json
        if os.path.exists(".secrets/huggingface_token.json"):
            with open(".secrets/huggingface_token.json", "r") as f:
                model_args["token"] = json.load(f)["token"]
    if "comet_experiment" not in model_args or model_args["comet_experiment"] is None:
        model_args["comet_experiment"] = f"{time_now}-{experiment.name}"
    # Train the model
    if experiment.training_settings.training_args.output_dir is None or experiment.training_settings.training_args.output_dir == "":
        experiment.training_settings.training_args.output_dir = os.path.join(model_log_dir, experiment.name)
    else:
        experiment.training_settings.training_args.output_dir = os.path.join(experiment.training_settings.training_args.output_dir, experiment.name)
    training_data.logger.info(f"Counting tokens in training dataset")
    training_data_formatter_callback = experiment.training_data_settings.training_data_formatter_type.get_class()
    total_tokens = token_count(experiment, hf_training_dataset, training_data_formatter_callback(), training_data.logger)
    training_data.logger.info(f"Total tokens in training dataset: {total_tokens}")


import typing
import random
import os
import tracemalloc
from datetime import datetime
try:
    from .best_first_search import BestFirstSearch
    from .beam_search import BeamSearch
    from .search import Node, Edge, SearchAlgorithm
except ImportError:
    from best_first_search import BestFirstSearch
    from beam_search import BeamSearch
    from search import Node, Edge, SearchAlgorithm

class Heuristic:
    def __init__(self, goal_value: int):
        self.goal_value = goal_value

    def __call__(self, node: Node) -> float:
        return abs(int(node.name) - self.goal_value)

class GenerateChildren:
    def __init__(self, start_value: int, goal_value: int, goal_level: int):
        self.goal_value = goal_value
        self.goal_level = goal_level
        self._start_node = Node(str(start_value))
        self._node_map = {}
        self._parent_map = {self._start_node: []}
        self._level = 0

    def __call__(self, node: Node) -> typing.Tuple[typing.List[Node], typing.List[Edge]]:
        """
        Generates a random number of child nodes for a given node. Each child node's value is
        either +1, +2, +3, +4, +5, .., +50 from the parent node's value, chosen randomly.
        """
        if node.name in self._node_map:
            new_node, _ = self._node_map[node.name]
            assert isinstance(new_node.children, list), "Children must be a list"
            return new_node.children, new_node.edges
        else:
            parent_value = int(node.name)
            num_children = random.randint(0, 30)  # Randomly choose how many children to generate (0 to 3 for this example)
            children = []
            edges = []
            unique_values = set()  # Keep track of unique child values
            for _ in range(num_children):
                increment = random.choice([i for i in range(51)])  # Choose either +0, +1, +2, +3, +4, .., +49
                child_value = parent_value + increment
                if child_value not in unique_values and child_value != self.goal_value:
                    unique_values.add(child_value)
                    children.append(Node(str(child_value)))
                    edges.append(Edge(str(increment), 0, None))
            parent_map = self._parent_map[node]
            self._level = 0 if len(parent_map) == 0 else max([self._node_map[parent.name][1] for parent in parent_map]) + 1
            if self._level == self.goal_level:
                children.append(Node(str(self.goal_value)))
                edges.append(Edge(str(self.goal_value - parent_value), 0, None))
            if self._level > self.goal_level:
                children = []
            if self._level != self.goal_level:
                assert all([int(child.name) != self.goal_value for child in children]), "Goal value should not be in children"
            new_node = Node(str(parent_value))
            cloned_children = [Node(child.name) for child in children]
            for child in cloned_children:
                new_node.add_child(child)
                if child not in self._parent_map:
                    self._parent_map[child] = [new_node]
                else:
                    self._parent_map[child].append(new_node)
            self._node_map[node.name] = (new_node, self._level)
            assert isinstance(children, list), "Children must be a list"
            return children, edges

def test_search_algorithm(algo: SearchAlgorithm, start_value: int = 0, goal_value: int = 10, goal_level: int = 3, timeout_in_secs: float = 60, generate_children: typing.Callable[[Node], typing.List[Node]] = None) -> typing.Callable[[Node], typing.List[Node]]:
    tracemalloc.start()
    current, peak = tracemalloc.get_traced_memory()
    # print("Current memory usage:", current / 10**6, "MB")
    # print("Peak memory usage:", peak / 10**6, "MB")
    start_memory = current
    start_peak = peak
    start_node = Node(str(start_value))
    goal_node = Node(str(goal_value))

    heuristic = Heuristic(goal_value)  # Generate the heuristic function based on the goal value
    # Generate the dynamic child generation function
    generate_children = GenerateChildren(start_value, goal_value, goal_level) if generate_children is None else generate_children 

    print("Testing {}...".format(algo.__class__.__name__))
    # Run the search algorithm with the generated heuristic and dynamic child generation
    goal_node_in_tree, found, time = algo.search(start_node, goal_node, heuristic, generate_children, 4, True, timeout_in_secs=timeout_in_secs)
    print("Search complete")
    print("Time elapsed:", time, "seconds")
    # print("Memory usage:", current / 10**6, "MB")
    # print("Peak memory usage:", peak / 10**6, "MB")
    # print("Memory usage change:", (current - start_memory) / 10**6, "MB")
    current, peak = tracemalloc.get_traced_memory()
    print("Peak memory usage change:", (peak - start_peak) / 10**6, "MB")
    tracemalloc.stop()
    if found:
        path = algo.reconstruct_path(start_node, goal_node_in_tree)
        # print("Path to goal:", [node.name for node in path])
        all_paths = algo.reconstruct_all_paths(start_node, goal_node_in_tree)
        if len(all_paths) > 1:
            print("Found multiple paths to goal:")
            # for path in all_paths:
            #     print([node.name for node in path])
        else:
            one_path = all_paths[0]
            assert len(path) == len(one_path), "Reconstructed path length does not match"
            assert all([node.name == path_node.name for node, path_node in zip(path, one_path)]), "Reconstructed path does not match"
            print("Only one path found to goal")
    else:
        all_paths = []
        print("No path found to goal")
    dot = algo.visualize_search(start_node, show=False, mark_paths=all_paths)
    print("Saving visualization...")
    os.makedirs(".log/search/test", exist_ok=True)
    time_now = datetime.now().strftime("%Y%m%d_%H%M%S")
    file_name = ".log/search/test/{}_{}".format(algo.__class__.__name__, time_now)
    dot.render(file_name, format='png', quiet=True)
    return generate_children # Return the dynamic child generation function for testing testing other algorithms

# Example usage
if __name__ == "__main__":
    random.seed(0xfacade)
    # See the memory usage of the search algorithms
    children_map = test_search_algorithm(BestFirstSearch(), 0, 1000, 12, timeout_in_secs=100)
    random.seed(0xfacade)
    test_search_algorithm(BeamSearch(2), 0, 1000, 12, timeout_in_secs=100, generate_children=children_map)


#!/usr/bin/env python3
import os
os.environ["COMET_MODE"] = "DISABLED"
os.environ["TOKENIZERS_PARALLELISM"] = "false"

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import hydra
import copy
import logging
import random
import time
import math
import typing
import ray
import multiprocessing
import unicodedata
import re
import uuid
import random
multiprocessing.set_start_method('spawn', force=True)
from datetime import datetime
from proof_wala.tools.ray_utils import clean_ray_logs
from proof_wala.llm_helpers.model import Model
from proof_wala.proof_search.search_driver import ProofSearchDriver
from proof_wala.proof_search.llm_tactic_generator import CodeT5PromptFormatter, CodeT5ResponseParser, LlmProofActionGenerator
from proof_wala.main.eval_config import EnvSettings, EvalBenchmark, EvalDataset, EvalFile, EvalProofResults, EvalProofResultsActor, EvalRunCheckpointInfoActor, EvalSettings, Experiments, EvalRunCheckpointInfo, parse_config
from itp_interface.tools.log_utils import setup_logger
from itp_interface.rl.proof_tree import ProofSearchResult
from itp_interface.rl.simple_proof_env import ProofEnv, ProofAction
from itp_interface.tools.proof_exec_callback import ProofExecutorCallback
from itp_interface.tools.dynamic_coq_proof_exec import DynamicProofExecutor as DynamicCoqProofExecutor
from itp_interface.tools.dynamic_lean_proof_exec import DynamicProofExecutor as DynamicLeanProofExecutor
from itp_interface.tools.lean4_sync_executor import get_all_theorems_in_file as get_all_theorems_in_file_lean4, get_fully_qualified_theorem_name as get_fully_qualified_theorem_name_lean4, get_theorem_name_resembling as get_theorem_name_resembling_lean4

def slugify(value, allow_unicode=False):
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize('NFKC', value)
    else:
        value = unicodedata.normalize('NFKD', value).encode('ascii', 'ignore').decode('ascii')
    value = re.sub(r'[^\w\s-]', '', value.lower())
    return re.sub(r'[-\s]+', '-', value).strip('-_')

def check_query_limit_reached(max_query_limit: int) -> typing.Callable[[int, typing.Dict[str, typing.Any]], bool]:
    def _check_query_limit_reached(steps: int, info: typing.Dict[str, typing.Any]):
        return info["queries"] >= max_query_limit
    return _check_query_limit_reached

def query_limit_info_message(max_query_limit: int) -> typing.Callable[[int, typing.Dict[str, typing.Any]], str]:
    def _query_limit_info_message(steps: int, info: typing.Dict[str, typing.Any]):
        return f"Step {info['queries']}/{max_query_limit} (Actual steps: {steps})"
    return _query_limit_info_message

def get_all_lemmas(coq_proof_exec_callback: ProofExecutorCallback, logger: logging.Logger):
    lemmas_to_prove = []
    if coq_proof_exec_callback.language == ProofAction.Language.LEAN4:
        lemmas_to_prove = get_all_theorems_in_file_lean4(coq_proof_exec_callback.file_path, use_cache=True)
        lemmas_to_prove = [get_fully_qualified_theorem_name_lean4(lemma) for lemma in lemmas_to_prove]
    else:
        with coq_proof_exec_callback.get_proof_executor() as main_executor:
            if isinstance(main_executor, DynamicLeanProofExecutor):
                main_executor.run_all_without_exec()
                lemmas_to_prove = main_executor.find_all_theorems_names()
            elif isinstance(main_executor, DynamicCoqProofExecutor):
                while not main_executor.execution_complete:
                    assert not main_executor.is_in_proof_mode(), "main_executor must not be in proof mode"
                    _ = list(main_executor.run_till_next_lemma_return_exec_stmt())
                    if main_executor.execution_complete:
                        break
                    lemma_name = main_executor.get_lemma_name_if_running()
                    if lemma_name is None:
                        _ = list(main_executor.run_to_finish_lemma_return_exec())
                        if main_executor.execution_complete:
                            break
                    else:
                        logger.info(f"Discovered lemma: {lemma_name}")
                        lemmas_to_prove.append(lemma_name)
                        main_executor.run_to_finish_lemma()
            else:
                raise Exception(f"Unsupported proof executor: {main_executor}")
    logger.info(f"Discovered {len(lemmas_to_prove)} lemmas")
    return lemmas_to_prove

class _Get_All_Lemmas:
    def __call__(self, ret_dict, path, get_all_lemmas_proof_exec_callback, logger: logging.Logger):
        try:
            ret_dict["lemmas"] = get_all_lemmas(get_all_lemmas_proof_exec_callback, logger)
        except:
            logger.exception(f"Exception occurred while getting all lemmas in file: {path}")

@ray.remote
def discover_lemmas_in_benchmark(
    skip_files_in_checkpoint: bool,
    eval_benchmark: EvalBenchmark, 
    dataset: EvalDataset, 
    eval_checkpoint_info: EvalRunCheckpointInfoActor,
    eval_settings: EvalSettings,
    eval_proof_results: EvalProofResultsActor,
    logfile: str):
    log_id = str(uuid.uuid4())
    logger = setup_logger(f"discover_lemmas_in_benchmark_{log_id}", logfile, format='%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    if eval_settings.do_lemmas_discovery:
        for file in dataset.files:
            path = os.path.join(dataset.project, file.path)
            theorem_maps : typing.Dict[str, typing.Dict[str, ProofSearchResult]] = ray.get(eval_proof_results.get_theorem_map.remote())
            if skip_files_in_checkpoint and path in theorem_maps:
                logger.info(f"Skipping the file: {path} as it was already attempted before.")
                continue
            ray.get(eval_checkpoint_info.add_path_to_maps.remote(path))
            ray.get(eval_proof_results.add_path_to_maps.remote(path))
            get_all_lemmas_proof_exec_callback = ProofExecutorCallback(
                project_folder=dataset.project,
                file_path=path,
                language=eval_benchmark.language,
                use_hammer=False, # We don't need hammer for this
                timeout_in_secs=eval_settings.timeout_in_secs,
                use_human_readable_proof_context=eval_settings.use_human_readable_proof_context,
                suppress_error_log=True,
                always_use_retrieval=False,
                logger=logger,
                setup_cmds=eval_benchmark.setup_cmds)
            manager = multiprocessing.Manager()
            return_dict = manager.dict()
            file_time_out = eval_settings.timeout_in_secs * eval_settings.max_proof_depth * 50
            logger.info(f"Getting all lemmas in file: {path} with timeout: {file_time_out} seconds")
            p = multiprocessing.Process(target=_Get_All_Lemmas(), args=(return_dict, path, get_all_lemmas_proof_exec_callback, logger))
            p.start()
            p.join(file_time_out)
            if p.is_alive():
                p.kill()
                p.join()
            p.close()
            if "lemmas" not in return_dict:
                logger.info(f"Failed to get all lemmas in file: {path}, moving on to the next file.")
                continue
            lemmas_to_prove = list(return_dict["lemmas"])
            if isinstance(file.theorems, str) and file.theorems == "*":
                file.theorems = list(lemmas_to_prove)
                file.theorems.sort() # sort to ensure one order when no theorems are specified
            elif isinstance(file.theorems, list):
                logger.info(f"Discovered {len(lemmas_to_prove)} lemmas in file: {path}")
                if get_all_lemmas_proof_exec_callback.language == ProofAction.Language.LEAN4:
                    logger.info("Converting theorems to fully qualified names for Lean 4.")
                    actual_theorem_name = [get_theorem_name_resembling_lean4(path, lemma, use_cache=True) for lemma in file.theorems]
                    for _lemma in actual_theorem_name:
                        logger.info(f"Converted lemma: {_lemma}")
                    for _lemma in lemmas_to_prove:
                        logger.info(f"Discovered lemma: {_lemma}")
                else:
                    actual_theorem_name = file.theorems
                if not dataset.negation:
                    # Check all theorems which can be proved
                    intersection = set(actual_theorem_name).intersection(lemmas_to_prove)
                    for _lemma in intersection:
                        logger.info(f"Intersection lemma: {_lemma}")
                    # Arrange them in the order of the file.theorems
                    file.theorems = [x for x in actual_theorem_name if x in intersection]
                else:
                    logger.info("Dataset is negation dataset. So, removing the theorems which are mentioned in the dataset.")
                    # Take the difference of the theorems which can be proved
                    difference = set(lemmas_to_prove).difference(file.theorems)
                    for lemma in difference:
                        logger.info(f"Lemma: {lemma} is not in the dataset.")
                    difference = list(difference)
                    difference.sort()
                    file.theorems = difference
            else:
                raise ValueError(f"Invalid theorems: {file.theorems}")
            logger.info(f"Discovered {len(file.theorems)} lemmas to prove in {path}")
            logger.info(f"Lemmas to prove in file {path}: \n{file.theorems}")
            if eval_settings.sample < 1.0:
                sample_size = math.ceil(len(file.theorems) * eval_settings.sample)
                logger.info(f"Sampling {sample_size} lemmas from {len(file.theorems)} lemmas in file {path}")
                file.theorems = list(random.sample(file.theorems, sample_size))
                logger.info(f"Sampled lemmas to prove in file {path}: \n{file.theorems}")
    else:
        logger.info("Skipping lemmas discovery")
    clean_ray_logs(logger)
    return dataset

def eval_dataset_once(
    model: Model, 
    proof_attempts_done: bool, 
    skip_files_in_checkpoint: bool, 
    track_time: bool,
    eval_benchmark: EvalBenchmark, 
    dataset: EvalDataset, 
    eval_checkpoint_info: EvalRunCheckpointInfoActor,
    eval_settings: EvalSettings,
    env_settings: EnvSettings,
    eval_proof_results: EvalProofResultsActor,
    time_budget_tracker: typing.Dict[str, typing.Dict[str, float]],
    attempt_idx: int,
    logger: logging.Logger,
    reshuffle: bool = False):
    if proof_attempts_done:
        return True
    any_proof_attempted = False
    if eval_settings.proof_tracer is not None:
        eval_settings.proof_tracer.reset_training_data(logger)
    if reshuffle:
        dataset_files = list(dataset.files)
        # Reshuffle the files
        random.shuffle(dataset_files)
    else:
        dataset_files = dataset.files
    current_attempt_idx = attempt_idx
    logger.info(f"Starting evaluation for dataset: {dataset.project}, attempt: {attempt_idx + 1}")
    for file in dataset_files:
        clean_ray_logs(logger)
        path = os.path.join(dataset.project, file.path)
        if track_time and path not in time_budget_tracker:
            if len(file.max_time_limits_in_secs) > 0:
                time_budget_tracker[path] = copy.deepcopy(file.max_time_limits_in_secs)
            else:
                time_budget_tracker[path] = {}
        proof_dump_file_name = os.path.join(eval_settings.proof_dump_dir, f"{path.replace('/', '_')}.txt")
        theorem_maps : typing.Dict[str, typing.Dict[str, ProofSearchResult]] = ray.get(eval_proof_results.get_theorem_map.remote())
        if skip_files_in_checkpoint and path in theorem_maps:
            logger.info(f"Skipping the file: {path} as it was already attempted before.")
            # The proof result for this file is already in the proof results
            # So we just log the proof result
            for lemma_name, proof_res_chkpt in theorem_maps[path].items():
                logger.info(f"Dumping proof search result:\n{proof_res_chkpt}")
                logger.info(f"Prover for lemma: {lemma_name} in file {path} completed.")
            continue
        if not os.path.exists(proof_dump_file_name):
            with open(proof_dump_file_name, "w") as f:
                f.write(f"File: {path}\n")
                f.write(f"Dataset:\n {dataset.to_json(indent=4)}\n")
                f.write(f"Evaluation Settings:\n {eval_settings.to_json(indent=4)}\n")
        ray.get(eval_checkpoint_info.add_path_to_maps.remote(path))
        ray.get(eval_proof_results.add_path_to_maps.remote(path))
        proof_exec_callback = ProofExecutorCallback(
            project_folder=dataset.project,
            file_path=path,
            language=eval_benchmark.language,
            use_hammer=eval_settings.use_hammer,
            timeout_in_secs=eval_settings.timeout_in_secs,
            use_human_readable_proof_context=eval_settings.use_human_readable_proof_context,
            suppress_error_log=True,
            always_use_retrieval=eval_settings.always_use_useful_theorem_retrieval,
            logger=logger,
            setup_cmds=eval_benchmark.setup_cmds,
            enable_search=False) # TODO: Make this configurable
        if reshuffle:
            file_theorems = list(file.theorems)
            random.shuffle(file_theorems)
        else:
            file_theorems = file.theorems
        for lemma_name in file_theorems:
            no_proof_res = ProofSearchResult(
                None, 
                False, 
                lemma_name, 
                [], 
                -1, 
                -1, 
                possible_failed_paths=-1, 
                num_of_backtracks=-1, 
                is_timeout=False, 
                is_inference_exhausted=False, 
                longest_success_path=-1,
                additional_info={},
                language=eval_benchmark.language)
            try:
                if track_time and lemma_name not in time_budget_tracker[path]:
                    time_budget_tracker[path][lemma_name] = eval_benchmark.timeout_per_theorem_in_secs
                if track_time and time_budget_tracker[path][lemma_name] <= 0:
                    logger.info(f"Time budget exhausted for lemma: {lemma_name} in file {path} so skipping it.")
                    continue
                logger.info(f"Will check if attempt possible for lemma: {lemma_name} [# {attempt_idx + 1}/{eval_settings.proof_retries}]")

                max_seq_length = eval_settings.max_seq_length
                character_per_token = eval_settings.character_per_token
                # TODO: Move no_steps to eval_settings
                prompt_formatter = CodeT5PromptFormatter(max_token_in_prompt=max_seq_length, character_per_token=character_per_token, no_steps=True)
                response_parser = CodeT5ResponseParser()
                width = eval_settings.width
                max_new_tokens=eval_settings.max_tokens_per_action
                temperature=eval_settings.temperature # Nucleus sampling
                do_sample=eval_settings.do_sample # Nucleus sampling
                top_k=eval_settings.top_k # Nucleus sampling
                stop_tokens=eval_settings.stop_tokens
                padding=eval_settings.padding
                return_full_text=eval_settings.return_full_text
                compute_probabilities=eval_settings.compute_probabilities
                # Get theorem_maps again because the previous attempt can be overwriiten
                theorem_maps: typing.Dict[str, typing.Dict[str, ProofSearchResult]] = ray.get(eval_proof_results.get_theorem_map.remote())
                proof_res_chkpt = theorem_maps.get(path, {}).get(lemma_name, None)
                max_retry_attempts = file.max_retry_attempts_limits.get(lemma_name, eval_settings.proof_retries)
                if proof_res_chkpt is None or (not proof_res_chkpt.proof_found and proof_res_chkpt.additional_info["attempt_idx"] < attempt_idx): #max_retry_attempts - 1):
                    any_proof_attempted = True
                    manager = multiprocessing.Manager()
                    return_dict = manager.dict()
                    def _run_prover(ret_dict):
                        try:
                            env = ProofEnv(f"basic_proof_env_{lemma_name}", 
                                proof_exec_callback, 
                                lemma_name, 
                                retrieval_strategy=env_settings.retrieval_strategy, 
                                max_proof_depth=eval_settings.max_proof_depth, 
                                always_retrieve_thms=eval_settings.always_use_useful_theorem_retrieval, 
                                logger=logger)
                            algo = eval_settings.get_search_algo()
                            generator = LlmProofActionGenerator(
                                model, prompt_formatter, response_parser, width, logger,
                                max_new_tokens=max_new_tokens, temperature=temperature, do_sample=do_sample, top_k=top_k,
                                stop_tokens=stop_tokens, padding=padding, 
                                return_full_text=return_full_text, 
                                compute_probabilities=compute_probabilities)
                            proof_search_heuristic = eval_settings.get_proof_search_heuristic()
                            search_driver = ProofSearchDriver(
                                algo,
                                generator,
                                proof_search_heuristic,
                                width=eval_settings.width,
                                logger=logger,
                                tracer=eval_settings.proof_tracer)
                            start_node, tree_node, proof_res = search_driver.search_proof(
                                env,
                                timeout_in_secs=eval_benchmark.timeout_per_theorem_in_secs)
                            logger.info("Proof search finished.")
                            if start_node is not None:
                                proof_paths = []
                                if proof_res.proof_found:
                                    proof_paths = algo.reconstruct_all_paths(start_node, tree_node)
                                dot = algo.visualize_search(start_node, show=False, mark_paths=proof_paths)
                                lemma_file_name = slugify(lemma_name, allow_unicode=False)
                                tree_dump_folder = os.path.join(eval_settings.proof_dump_dir, "proof_trees", lemma_file_name)
                                try:
                                    # if lemma_file_name can be a valid file name
                                    os.makedirs(tree_dump_folder, exist_ok=True)
                                except:
                                    logger.exception(f"Failed to create the proof tree dump folder for lemma: {lemma_name} in file {path}")
                                    lemma_name_hash = hash(lemma_name)
                                    lemma_file_name = lemma_name_hash
                                    tree_dump_folder = os.path.join(eval_settings.proof_dump_dir, "proof_trees", lemma_file_name)
                                    os.makedirs(tree_dump_folder, exist_ok=True)
                                    logger.info("\nUsing hash of lemma name:" +
                                        f"\n {lemma_name} to create the folder for dumping the proof tree in file {file.path}" +
                                        f"\n The hash is: {lemma_name_hash}")
                                time_stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
                                tree_dump_path = os.path.join(tree_dump_folder, f"{lemma_file_name}_{time_stamp}")
                                dot.render(tree_dump_path, format='svg', quiet=True)
                            ret_dict["proof_res"] = proof_res
                            ret_dict["attempted_success"] = True
                            ret_dict["service_down"] = False
                            additional_info = proof_res_chkpt.additional_info if proof_res_chkpt is not None else {'attempt_idx': 0}
                            proof_res.additional_info = additional_info
                            logger.info(f"Dumping proof search result:\n{proof_res}")
                            logger.info(f"Prover for lemma: {lemma_name} in file {path} completed.")
                        except:
                            logger.exception(f"Exception occurred while proving lemma: {lemma_name} in file {path}")
                            ret_dict["attempted_success"] = False
                            ret_dict["service_down"] = False

                    should_retry = True
                    max_retry = 4 # This retry is only when for some mysterious reason the llama service goes down
                    if proof_res_chkpt is not None:
                        current_attempt_idx = proof_res_chkpt.additional_info["attempt_idx"]
                        logger.info(f"Previous attempt for proving lemma: {lemma_name} in file {path} was: {attempt_idx}")
                        logger.info(f"Previous proof search result:\n{proof_res_chkpt}")
                    logger.info(f"Attempt {attempt_idx + 1 if proof_res_chkpt is None else current_attempt_idx + 2} for proving lemma: {lemma_name} in file {path}")
                    while should_retry and max_retry > 0:
                        # Run the prover with a timeout
                        timeout = min(eval_settings.timeout_in_secs * eval_settings.max_proof_depth * 1.25, eval_benchmark.timeout_per_theorem_in_secs)
                        if track_time and time_budget_tracker[path][lemma_name] < timeout:
                            timeout = time_budget_tracker[path][lemma_name]
                        logger.info(f"Running the prover agent for lemma: {lemma_name} with timeout: {timeout} seconds")
                        tic_start = time.time()
                        _run_prover(return_dict)
                        toc_end = time.time()
                        if track_time:
                            time_budget_tracker[path][lemma_name] -= (toc_end - tic_start)
                        if track_time and time_budget_tracker[path][lemma_name] <= 0:
                            logger.info(f"Time budget exhausted for lemma: {lemma_name} in file {path}")
                            proof_attempt_idx = (proof_res_chkpt.additional_info["attempt_idx"] + 1) if proof_res_chkpt is not None and "attempt_idx" in proof_res_chkpt.additional_info else current_attempt_idx
                            proof_res_chkpt = copy.deepcopy(no_proof_res)
                            proof_res_chkpt.is_timeout = True
                            proof_res_chkpt.proof_time_in_secs = toc_end - tic_start
                            proof_res_chkpt.additional_info["attempt_idx"] = proof_attempt_idx
                            ray.get(eval_proof_results.add_theorem_to_maps.remote(path, lemma_name, proof_res_chkpt))
                            ray.get(eval_checkpoint_info.add_theorem_to_maps.remote(path, lemma_name, False))
                            should_retry = False
                        elif "attempted_success" not in return_dict:
                            logger.info(f"Prover Agent for lemma: {lemma_name} in file {path} got killed as it timed out.")
                            proof_attempt_idx = (proof_res_chkpt.additional_info["attempt_idx"] + 1) if proof_res_chkpt is not None and "attempt_idx" in proof_res_chkpt.additional_info else current_attempt_idx
                            proof_res_chkpt = copy.deepcopy(no_proof_res)
                            proof_res_chkpt.is_timeout = True
                            proof_res_chkpt.proof_time_in_secs = toc_end - tic_start
                            proof_res_chkpt.additional_info["attempt_idx"] = proof_attempt_idx
                            ray.get(eval_proof_results.add_theorem_to_maps.remote(path, lemma_name, proof_res_chkpt))
                            ray.get(eval_checkpoint_info.add_theorem_to_maps.remote(path, lemma_name, False))
                            should_retry = False
                        elif not return_dict["attempted_success"]:
                            if not return_dict["service_down"] or \
                                (eval_settings.model_name is not None and \
                                len(eval_settings.model_name) != 0) or \
                                max_retry <= 1:
                                logger.info(f"Failed to prove lemma: {lemma_name} in file {path}")
                                proof_attempt_idx = (proof_res_chkpt.additional_info["attempt_idx"] + 1) if proof_res_chkpt is not None and "attempt_idx" in proof_res_chkpt.additional_info else current_attempt_idx
                                proof_res_chkpt = copy.deepcopy(no_proof_res)
                                proof_res_chkpt.is_timeout = True
                                proof_res_chkpt.proof_time_in_secs = toc_end - tic_start
                                proof_res_chkpt.additional_info["attempt_idx"] = proof_attempt_idx
                                ray.get(eval_proof_results.add_theorem_to_maps.remote(path, lemma_name, proof_res_chkpt))
                                ray.get(eval_checkpoint_info.add_theorem_to_maps.remote(path, lemma_name, False))
                                should_retry = False
                        else:
                            logger.info(f"Prover for lemma: {lemma_name} in file {path} completed.")
                            proof_attempt_idx = (proof_res_chkpt.additional_info["attempt_idx"] + 1) if proof_res_chkpt is not None and "attempt_idx" in proof_res_chkpt.additional_info else current_attempt_idx
                            proof_res_chkpt : ProofSearchResult = return_dict["proof_res"]
                            proof_res_chkpt.additional_info["attempt_idx"] = proof_attempt_idx
                            ray.get(eval_proof_results.add_theorem_to_maps.remote(path, lemma_name, proof_res_chkpt))
                            ray.get(eval_checkpoint_info.add_theorem_to_maps.remote(path, lemma_name, True))
                            should_retry = False
                        return_dict.clear()
                        max_retry -= 1
                else:
                    proof_res_attempt_idx = proof_res_chkpt.additional_info["attempt_idx"]
                    if proof_res_attempt_idx == attempt_idx:
                        logger.info(f"Dumping proof search result:\n{proof_res_chkpt}")
                        logger.info(f"Prover for lemma: {lemma_name} in file {path} completed.")
                    else:
                        logger.info(f"Skipping the attempt for proving lemma: {lemma_name} in file {path} as it was already attempted before.")
            except:
                logger.exception(f"Exception occurred while proving lemma: {lemma_name} in file {path}")
                proof_res_chkpt = copy.deepcopy(no_proof_res)
                proof_res_chkpt.is_timeout = True
                proof_res_chkpt.additional_info["attempt_idx"] = attempt_idx
                proof_res_chkpt.additional_info["total_queries"] = 0
                proof_res_chkpt.proof_time_in_secs = 0
                ray.get(eval_proof_results.add_theorem_to_maps.remote(path, lemma_name, proof_res_chkpt))
                ray.get(eval_checkpoint_info.add_theorem_to_maps.remote(path, lemma_name, False))
                # eval_proof_results.add_theorem_to_maps(path, lemma_name, proof_res_chkpt)
                # eval_checkpoint_info.add_theorem_to_maps(path, lemma_name, False)
    
    if eval_settings.proof_tracer is not None:
        eval_settings.proof_tracer.save()
    proof_attempts_done = not any_proof_attempted
    return proof_attempts_done

@ray.remote
def eval_dataset_once_with_retries(
    time_budget_tracker: typing.Dict[str, typing.Dict[str, float]],
    attempt_idx: int,
    model_path: str,
    is_seq2seq: bool,
    proof_attempts_done: bool, 
    skip_files_in_checkpoint: bool, 
    track_time: bool,
    eval_benchmark: EvalBenchmark, 
    dataset: EvalDataset, 
    eval_checkpoint_info: EvalRunCheckpointInfoActor,
    eval_settings: EvalSettings,
    env_settings: EnvSettings,
    eval_proof_results: EvalProofResultsActor,
    log_dir: str,
    device_id: typing.Optional[int] = None):
    # time_budget_tracker = {}
    if device_id is not None:
        os.environ["CUDA_VISIBLE_DEVICES"] = str(device_id)
    model = Model(model_path, is_seq2seq=is_seq2seq, use_lora=False)
    model.__enter__()
    logfile = os.path.join(log_dir, f"eval_dataset_multiple_attempts.log")
    logger_id = str(uuid.uuid4())
    logger = setup_logger(f"eval_dataset_multiple_attempts_{logger_id}", logfile, format='%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    retry_cnt = 0
    reshuffle = False
    attempt_success = False
    eval_result = False
    while retry_cnt < 100 and not attempt_success:
        try:
            logger.info(f"(Retry Number #: {retry_cnt + 1}) Restarting evaluation for dataset: {dataset.project}")
            eval_result = eval_dataset_once(
                model,
                proof_attempts_done,
                skip_files_in_checkpoint,
                track_time,
                eval_benchmark,
                dataset,
                eval_checkpoint_info,
                eval_settings,
                env_settings,
                eval_proof_results,
                time_budget_tracker,
                attempt_idx,
                logger,
                reshuffle=reshuffle)
            attempt_success = True
        except:
            logger.exception(f"Exception occurred while evaluating dataset: {dataset.project}")
            if retry_cnt < 98:
                logger.info(f"Next Retry Count: {retry_cnt + 2}")
            else:
                logger.info(f"Final Retry Count: {retry_cnt + 2}")
            attempt_success = False
            reshuffle = True
            random_seed = random.randint(0, 1000000)
            logger.info(f"Setting random seed to: {random_seed}")
            random.seed(random_seed)
        if eval_result:
            break
        retry_cnt += 1
    return eval_result

@ray.remote
def eval_dataset_multiple_attempts(
    model_path: str,
    is_seq2seq: bool,
    proof_attempts_done: bool, 
    skip_files_in_checkpoint: bool, 
    track_time: bool,
    eval_benchmark: EvalBenchmark, 
    dataset: EvalDataset, 
    eval_checkpoint_info: EvalRunCheckpointInfoActor,
    eval_settings: EvalSettings,
    env_settings: EnvSettings,
    eval_proof_results: EvalProofResultsActor,
    log_dir: str,
    device_id: typing.Optional[int] = None):
    time_budget_tracker = {}
    # if device_id is not None:
    #     os.environ["CUDA_VISIBLE_DEVICES"] = str(device_id)
    # model = Model(model_path, is_seq2seq=is_seq2seq, use_lora=False)
    # model.__enter__()
    logfile = os.path.join(log_dir, f"eval_dataset_multiple_attempts.log")
    logger = setup_logger("eval_dataset_multiple_attempts", logfile, format='%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    logger.info(f"Starting evaluation for dataset: {dataset.project}")
    for file in dataset.files:
        logger.info(f"File in dataset: {file.path}")
        logger.info(f"Theorems in file: {file.theorems}")
    for attempt_idx in range(eval_settings.proof_retries):
        proof_attempts_done_remote = eval_dataset_once_with_retries.remote(
            time_budget_tracker,
            attempt_idx,
            model_path,
            is_seq2seq,
            proof_attempts_done, 
            skip_files_in_checkpoint, 
            track_time,
            eval_benchmark, 
            dataset, 
            eval_checkpoint_info,
            eval_settings,
            env_settings,
            eval_proof_results,
            log_dir,
            device_id)
        proof_attempts_done = ray.get(proof_attempts_done_remote)
        proof_attempts_done = proof_attempts_done and attempt_idx == eval_settings.proof_retries - 1
        if proof_attempts_done:
            logger.info('=' * 80)
            logger.info(f"Finished all proof attempts for dataset: {dataset.project}")
            logger.info('=' * 80)
        else:
            logger.info('=' * 80)
            logger.info(f"Proof attempt (# {attempt_idx + 1}/{eval_settings.proof_retries}) done for dataset: {dataset.project}")
            logger.info('=' * 80)

def eval_dataset(env_settings: EnvSettings, eval_benchmark: EvalBenchmark, dataset: EvalDataset, eval_settings: EvalSettings, eval_checkpoint_info: EvalRunCheckpointInfo, eval_proof_results: EvalProofResults, logger: logging.Logger = None):
    logger = logger if logger else logging.getLogger(__name__)
    skip_files_in_checkpoint = False if "SKIP_FILES_IN_CHECKPOINT" not in os.environ else bool(os.environ["SKIP_FILES_IN_CHECKPOINT"])
    follow_seed = False if "FOLLOW_SEED" not in os.environ else bool(os.environ["FOLLOW_SEED"])

    if eval_settings.proof_retries > 1:
        assert eval_settings.temperature > 0.0, "Proof retries is only supported for temperature > 0.0"

    proof_attempts_done = False
    if "STRICT_TIME_BUDGET_ACCROSS_ATTEMPTS" in os.environ and bool(os.environ["STRICT_TIME_BUDGET_ACCROSS_ATTEMPTS"]):
        track_time = True
        logger.info(f"Strict time budget across attempts is enabled. Proofs will not be attempted beyond {eval_benchmark.timeout_per_theorem_in_secs} seconds.")
    else:
        track_time = False
    model_path = eval_settings.model_name
    is_seq2seq = eval_settings.is_seq2seq
    # get CUDA_VISIBLE_DEVICES
    gpus = []
    if "CUDA_VISIBLE_DEVICES" in os.environ:
        num_gpus = len(os.environ["CUDA_VISIBLE_DEVICES"].split(","))
        max_model_parallelism = min(num_gpus, eval_settings.model_parallelism)
        gpus = os.environ["CUDA_VISIBLE_DEVICES"].split(",")
    else:
        max_model_parallelism = 1
    # Split the datset into max_model_parallelism parts
    dataset_chunks = []
    files_per_chunk = math.ceil(len(dataset.files) / max_model_parallelism)
    for i in range(max_model_parallelism):
        start = i * files_per_chunk
        end = min((i + 1) * files_per_chunk, len(dataset.files))
        files = dataset.files[start:end]
        if len(files) == 0:
            break
        dataset_chunks.append(EvalDataset(project=dataset.project, files=dataset.files[start:end], negation=dataset.negation))
    remotes = []
    # Add proof_wala to python path
    root_dir = f"{__file__.split('proof_wala')[0]}"
    os.environ["PYTHONPATH"] = f"{root_dir}:{os.environ.get('PYTHONPATH', '')}"
    eval_checkpoint_info_actor = EvalRunCheckpointInfoActor.remote(eval_checkpoint_info)
    eval_proof_results_actor = EvalProofResultsActor.remote(eval_proof_results)
    if follow_seed:
        random.seed(eval_settings.sample_seed)
    # First discover all the lemmas that has to be proved
    for i, dataset in enumerate(dataset_chunks):
        logdir = eval_checkpoint_info.logging_dirs[-1]
        logfile = os.path.join(logdir, f"discover_lemmas_in_benchmark_{i}.log")
        remotes.append(discover_lemmas_in_benchmark.remote(
            skip_files_in_checkpoint,
            eval_benchmark,
            dataset,
            eval_checkpoint_info_actor,
            eval_settings,
            eval_proof_results_actor,
            logfile))
    discovered_dataset_chunks : typing.List[EvalDataset] = ray.get(remotes)
    # No negation datasets are allowed now as we have already discovered the lemmas
    new_dataset_chunks = [EvalDataset(project=dataset.project, files=[], negation=False) for _ in range(max_model_parallelism)]
    # Redistribute the lemmas to the dataset chunks
    new_dataset_chunk_idx = 0
    new_dataset_chunk_path = [{} for _ in range(max_model_parallelism)]
    if follow_seed:
        seed = eval_settings.sample_seed
    else:
        seed = int(time.time()) % 1000000
    logger.info(f"Shuffling the discovered lemmas with seed: {seed}")
    random.seed(seed)
    random.shuffle(discovered_dataset_chunks)
    for i in range(len(discovered_dataset_chunks)):
        files = list(discovered_dataset_chunks[i].files)
        random.shuffle(files)
        for file in files:
            assert isinstance(file.theorems, list), f"Invalid theorems: {file.theorems}"
            thm_lst = list(file.theorems)
            random.shuffle(thm_lst)
            for thm in thm_lst:
                if file.path not in new_dataset_chunk_path[new_dataset_chunk_idx]:
                    new_file = EvalFile(file.path, [thm], 
                        file.max_retry_attempts_limits, 
                        file.max_time_limits_in_secs)
                    new_dataset_chunks[new_dataset_chunk_idx].files.append(new_file)
                    new_dataset_chunk_path[new_dataset_chunk_idx][file.path] = len(new_dataset_chunks[new_dataset_chunk_idx].files) - 1
                else:
                    new_file_idx = new_dataset_chunk_path[new_dataset_chunk_idx][file.path]
                    new_dataset_chunks[new_dataset_chunk_idx].files[new_file_idx].theorems.append(thm)
                new_dataset_chunk_idx = (new_dataset_chunk_idx + 1) % max_model_parallelism
    dataset_chunks = new_dataset_chunks

    for i, dataset in enumerate(dataset_chunks):
        gpu_id = gpus[i] if i < len(gpus) else None
        if gpu_id is not None:
            os.environ["CUDA_VISIBLE_DEVICES"] = gpu_id
        log_dir = eval_checkpoint_info.logging_dirs[-1]
        log_dir = os.path.join(log_dir, f"gpu_{gpu_id}_{i}")
        os.makedirs(log_dir, exist_ok=True)
        remotes.append(eval_dataset_multiple_attempts.remote(
            model_path,
            is_seq2seq,
            proof_attempts_done,
            skip_files_in_checkpoint,
            track_time,
            eval_benchmark,
            dataset,
            eval_checkpoint_info_actor,
            eval_settings,
            env_settings,
            eval_proof_results_actor,
            log_dir,
            gpu_id))
    ray.get(remotes)
    eval_checkpoint_info = ray.get(eval_checkpoint_info_actor.get_checkpoint_info.remote())
    eval_proof_results = ray.get(eval_proof_results_actor.get_proof_results.remote())
    return eval_checkpoint_info, eval_proof_results

def measure_success(benchmark : EvalBenchmark, eval_settings : EvalSettings, eval_proof_results: EvalProofResults, logger: logging.Logger = None):
    success_count = 0
    proofs_dump_file = os.path.join(eval_settings.proof_dump_dir, "benchmark_proof_results.txt")
    proof_dump_file_exists = os.path.exists(proofs_dump_file)
    open_mode = "a" if proof_dump_file_exists else "w"
    with open(proofs_dump_file, open_mode) as f:
        if not proof_dump_file_exists:
            f.write(f"Settings: \n{eval_settings.to_json(indent=4)}\n")
            f.write(f"Benchmark: \n{benchmark.to_json(indent=4)}\n")
        for path, proofs in eval_proof_results.theorem_map.items():
            for lemma_name, proof_res in proofs.items():
                if proof_res.proof_found:
                    success_count += 1
                    logger.info(f"Proof found for lemma: {lemma_name} in file {path}")
                else:
                    logger.info(f"Proof not found for lemma: {lemma_name} in file {path}")
                f.write(f"Lemma: {lemma_name}\n")
                f.write(f"File: {path}\n")
                f.write(f"Proof/Incomplete proof: \n{proof_res}\n")
        total_attempted = sum([len(x) for _, x in eval_proof_results.theorem_map.items()])
        logger.info(f"Success rate: {success_count}/{total_attempted} = {success_count/total_attempted} for benchmark: {benchmark.name}")
        f.write(f"Success rate: {success_count}/{total_attempted} = {success_count/total_attempted} for benchmark: {benchmark.name}\n")

def eval_benchmark(experiment: Experiments, log_dir: str, logger: logging.Logger = None):
    trial_cnt = 1
    env_settings = experiment.env_settings
    eval_settings = experiment.eval_settings
    benchmark = experiment.benchmark
    checkpoint_dir = experiment.eval_settings.checkpoint_dir
    eval_settings.checkpoint_dir = os.path.join(checkpoint_dir, benchmark.name, eval_settings.name)
    os.makedirs(eval_settings.checkpoint_dir, exist_ok=True)
    # Load the checkpoint file if it exists
    checkpoint_file = os.path.join(eval_settings.checkpoint_dir, "checkpoint_info.json")
    if os.path.exists(checkpoint_file):
        with open(checkpoint_file, "r") as f:
            checkpoint_info: EvalRunCheckpointInfo = EvalRunCheckpointInfo.from_json(f.read())
        eval_settings.proof_dump_dir = checkpoint_info.proof_dump_dir
        checkpoint_info.logging_dirs.append(log_dir)
    else:
        time_now = time.strftime("%Y%m%d-%H%M%S")
        eval_settings.proof_dump_dir = os.path.join(eval_settings.proof_dump_dir, benchmark.name, time_now)
        os.makedirs(eval_settings.proof_dump_dir, exist_ok=True)
        checkpoint_info = EvalRunCheckpointInfo(
            checkpoint_file=checkpoint_file,
            proof_dump_dir=eval_settings.proof_dump_dir, 
            logging_dirs=[log_dir], 
            theorem_maps={})
    eval_proof_file = os.path.join(eval_settings.proof_dump_dir, "proof_results.json")
    if os.path.exists(eval_proof_file):
        with open(eval_proof_file, "r") as f:
            eval_proof_results: EvalProofResults = EvalProofResults.from_json(f.read())
    else:
        eval_proof_results = EvalProofResults(
            path=os.path.join(eval_settings.proof_dump_dir, "proof_results.json"),
            theorem_map={})
    while trial_cnt > 0:
        try:
            logger = logger or logging.getLogger(__name__)
            for dataset in benchmark.datasets:
                checkpoint_info, eval_proof_results = eval_dataset(env_settings, benchmark, dataset, eval_settings, checkpoint_info, eval_proof_results, logger=logger)
            measure_success(benchmark, eval_settings, eval_proof_results, logger=logger)
            trial_cnt = 0
        except:
            trial_cnt -= 1
            logger.exception(f"Exception occurred. Retrying {trial_cnt} more times.")
            if trial_cnt <= 0:
                raise
            else:
                time.sleep(10)
    logger.info(f"Finished running experiment: \n{experiment.to_json(indent=4)}")

@hydra.main(config_path="config", config_name="eval_simple_lean_test_multilingual", version_base="1.2")
def main(cfg):
    experiment = parse_config(cfg)
    log_dif_prefix = ".log/evals/benchmark" if "log_dir" not in cfg else cfg.log_dir
    log_dir = os.path.join(log_dif_prefix, experiment.benchmark.name, time.strftime("%Y%m%d-%H%M%S"))
    # log_dir = ".log/evals/benchmark/{}/{}".format(experiment.benchmark.name, time.strftime("%Y%m%d-%H%M%S"))
    os.makedirs(log_dir, exist_ok=True)
    log_path = os.path.join(log_dir, "eval.log")
    logger = setup_logger(__name__, log_path, logging.INFO, format='%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s')
    logger.info(f"Pid: {os.getpid()}")
    logger.info(f"Running Experiment: {experiment.to_json(indent=4)}")
    eval_benchmark(experiment, log_dir, logger=logger)
    pass

def run_proof_search():
    # RayUtils.init_ray(num_of_cpus=20, object_store_memory_in_gb=50, memory_in_gb=1)
    # Start the ray cluster
    os.environ["PYTHONPATH"] = f"{root_dir}:{os.environ.get('PYTHONPATH', '')}"
    from filelock import FileLock
    import json
    os.makedirs(".log/locks", exist_ok=True)
    ray_was_started = False
    print("Starting run_proof_search Pid: ", os.getpid())
    with FileLock(".log/locks/ray.lock"):
        if os.path.exists(".log/ray/session_latest"):
            with open(".log/ray/session_latest", "r") as f:
                ray_session = f.read()
                ray_session = json.loads(ray_session)
            ray_address = ray_session["address"]
            ray.init(address=ray_address)
            print("Ray was already started")
        else:
            ray_session = ray.init(
                num_cpus=8, 
                object_store_memory=150*2**30, 
                _memory=150*2**30, 
                logging_level=logging.CRITICAL, 
                ignore_reinit_error=False, 
                log_to_driver=False, 
                configure_logging=False,
                _system_config={"metrics_report_interval_ms": 3*10**8})
            with open(".log/ray/session_latest", "w") as f:
                f.write(json.dumps(ray_session))
            ray_was_started = True
            print("Ray was started")
            print("Ray session: ", ray_session)
    main()

if __name__ == "__main__":
    run_proof_search()


#!/usr/bin/env python3

import torch
import torch.nn as nn
import typing
from logging import Logger
from torch.utils.data import DataLoader, Dataset
from typing import List, Union, Optional, Callable, Dict, Tuple
from transformers import PreTrainedModel, TrainingArguments, PreTrainedTokenizerBase, TrainerCallback, DataCollator
from transformers.trainer_utils import EvalLoopOutput, EvalPrediction, denumpify_detensorize, find_executable_batch_size
from trl import SFTTrainer
from transformers.trainer_callback import TrainerState
from transformers.trainer import logger as hf_trainer_logger
from transformers.integrations.deepspeed import deepspeed_init
from transformers import DataCollator

class GenerateEvalSFTTrainer(SFTTrainer):
    def __init__(
        self,
        model: Union[PreTrainedModel, nn.Module, str] = None,
        args: TrainingArguments = None,
        data_collator: Optional[DataCollator] = None,
        train_dataset: Optional[Dataset] = None,
        eval_dataset: Optional[Union[Dataset, Dict[str, Dataset]]] = None,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        model_init: Optional[Callable[[], PreTrainedModel]] = None,
        compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
        callbacks: Optional[List[TrainerCallback]] = None,
        optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (None, None),
        preprocess_logits_for_metrics: Optional[Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
        peft_config: Optional["PeftConfig"] = None,
        dataset_text_field: Optional[str] = None,
        packing: Optional[bool] = False,
        formatting_func: Optional[Callable] = None,
        max_seq_length: Optional[int] = None,
        infinite: Optional[bool] = None,
        num_of_sequences: Optional[int] = 1024,
        chars_per_token: Optional[float] = 3.6,
        dataset_num_proc: Optional[int] = None,
        dataset_batch_size: int = 1000,
        neftune_noise_alpha: Optional[float] = None,
        model_init_kwargs: Optional[Dict] = None,
        dataset_kwargs: Optional[Dict] = None,
        generative_compute_metrics: Optional[Callable[[str, TrainerState, List[str], List[str], List[List[str]]], Dict]] = None,
        generate_callback: Optional[Callable[[Dataset], Tuple[List[str], List[str], List[List[str]]]]] = None,
        logger: Logger = None,
    ):
        assert (generative_compute_metrics is None) == (generate_callback is None), "Both generative_compute_metrics and generate_callback must be defined or not defined"
        args.prediction_loss_only = True
        args.remove_unused_columns = False
        args.include_inputs_for_metrics = True
        super().__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            model_init=model_init,
            compute_metrics=compute_metrics,
            callbacks=callbacks,
            optimizers=optimizers,
            preprocess_logits_for_metrics=preprocess_logits_for_metrics,
            peft_config=peft_config,
            dataset_text_field=dataset_text_field,
            packing=packing,
            formatting_func=formatting_func,
            max_seq_length=max_seq_length,
            infinite=infinite,
            num_of_sequences=num_of_sequences,
            chars_per_token=chars_per_token,
            dataset_num_proc=dataset_num_proc,
            dataset_batch_size=dataset_batch_size,
            neftune_noise_alpha=neftune_noise_alpha,
            model_init_kwargs=model_init_kwargs,
            dataset_kwargs=dataset_kwargs
        )
        self.original_training_dataset = train_dataset
        self.original_eval_dataset = eval_dataset
        self.generative_compute_metrics = generative_compute_metrics
        self.generate_callback = generate_callback
        self.generate_batch_size = self.args.eval_batch_size
        self.logger = logger if logger is not None else hf_trainer_logger
        self._agg_metrics = {}
        self._num_evals = 1 if type(eval_dataset) is not dict else len(eval_dataset)
        self._metric_agg_idx = 0
        pass

    def get_eval_dataloader(self, eval_dataset: typing.Union[Dataset, None] = None) -> DataLoader:
        return None

    def evaluation_loop(self, 
            dataloader: DataLoader, 
            description: str, 
            prediction_loss_only: typing.Union[bool, None] = None, 
            ignore_keys: typing.Union[List[str], None] = None, 
            metric_key_prefix: str = "eval") -> EvalLoopOutput:
        """
        Prediction/evaluation loop, shared by `Trainer.evaluate()` and `Trainer.predict()`.

        Works both with or without labels.
        """
        args = self.args

        # NEW: no logits are returned, so we need to change the code here
        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else args.prediction_loss_only

        # if eval is called w/o train, handle model prep here
        if self.is_deepspeed_enabled and self.deepspeed is None:
            _, _ = deepspeed_init(self, num_training_steps=0, inference=True)

        model = self._wrap_model(self.model, training=False, dataloader=dataloader)

        if len(self.accelerator._models) == 0 and model is self.model:
            model = (
                self.accelerator.prepare(model)
                if self.is_deepspeed_enabled
                else self.accelerator.prepare_model(model, evaluation_mode=True)
            )

            if self.is_fsdp_enabled:
                self.model = model

            # for the rest of this function `model` is the outside model, whether it was wrapped or not
            if model is not self.model:
                self.model_wrapped = model

            # backward compatibility
            if self.is_deepspeed_enabled:
                self.deepspeed = self.model_wrapped

        # if full fp16 or bf16 eval is wanted and this ``evaluation`` or ``predict`` isn't called
        # while ``train`` is running, cast it to the right dtype first and then put on device
        if not self.is_in_train:
            if args.fp16_full_eval:
                model = model.to(dtype=torch.float16, device=args.device)
            elif args.bf16_full_eval:
                model = model.to(dtype=torch.bfloat16, device=args.device)

        model.eval()

        self.callback_handler.eval_dataloader = dataloader
        # Do this before wrapping.
        if args.past_index >= 0:
            self._past = None


        if args.past_index and hasattr(self, "_past"):
            # Clean the state at the end of the evaluation loop
            delattr(self, "_past")

        all_preds : List[List[str]] = []
        all_labels : List[str] = []
        all_inputs : List[str] = []
        if self.generate_callback is not None:
            if isinstance(self.original_eval_dataset, dict):
                assert metric_key_prefix.startswith("eval_")
                dataset_name = metric_key_prefix[len("eval_"):]
                original_eval_dataset = self.original_eval_dataset[dataset_name]
            else:
                assert metric_key_prefix == "eval"
                dataset_name = "eval"
                original_eval_dataset = self.original_eval_dataset
            _eval_cnt = 0
            def _eval_callback(batch_size: int):
                self.logger.info(f"***** Running {description} *****")
                self.logger.info(f"  Batch size = {batch_size}")
                nonlocal _eval_cnt
                current_eval_cnt = 0
                dataloader_params = {
                    "batch_size": batch_size,
                    "num_workers": self.args.dataloader_num_workers,
                    "pin_memory": self.args.dataloader_pin_memory,
                }
                if not isinstance(original_eval_dataset, torch.utils.data.IterableDataset):
                    dataloader_params["drop_last"] = self.args.dataloader_drop_last
                new_dataloader = DataLoader(original_eval_dataset, **dataloader_params)
                for step, inputs in enumerate(new_dataloader):
                    if current_eval_cnt + len(inputs) < _eval_cnt:
                        current_eval_cnt += len(inputs)
                        continue
                    try:
                        prompts, completions, generated = self.generate_callback(inputs)
                        all_inputs.extend(prompts)
                        all_labels.extend(completions)
                        all_preds.extend(generated)
                        _eval_cnt += len(prompts)
                        current_eval_cnt += len(inputs)
                        self.logger.info(f"Evaluated: {_eval_cnt} / {len(original_eval_dataset)}")
                    except Exception as e:
                        self.logger.error(f"Error evaluating: {e}")
                        self.logger.error(f"Skipping: {_eval_cnt} / {len(original_eval_dataset)}")
                return batch_size
            generation_evaluation = find_executable_batch_size(_eval_callback, starting_batch_size=self.generate_batch_size, auto_find_batch_size=self.args.auto_find_batch_size)
            self.generate_batch_size = generation_evaluation()

        # Metrics!
        if self.generative_compute_metrics is not None:
            assert self.generate_callback is not None
            if args.include_inputs_for_metrics:
                metrics = self.generative_compute_metrics(dataset_name, self.state, all_inputs, all_labels, all_preds)
        else:
            metrics = {}

        # To be JSON-serializable, we need to remove numpy types or zero-d tensors
        metrics = denumpify_detensorize(metrics)

        if hasattr(self, "jit_compilation_time"):
            metrics[f"{metric_key_prefix}_jit_compilation_time"] = self.jit_compilation_time

        # Prefix all keys with metric_key_prefix + '_'
        for key in list(metrics.keys()):
            if self._num_evals >= 1:
                if self._metric_agg_idx == 0:
                    self._agg_metrics[key] = 0
                    self._agg_metrics["eval_count"] = 0
            if not key.startswith(f"{metric_key_prefix}_"):
                metrics[f"{metric_key_prefix}_{key}"] = metrics.pop(key)
                if self._num_evals >= 1:
                    if self._agg_metrics["eval_count"] == 0 and _eval_cnt == 0:
                        self._agg_metrics["eval_count"] = 1e-6 # Avoid division by zero
                    self._agg_metrics[key] = (self._agg_metrics[key] * self._agg_metrics['eval_count'] + metrics[f"{metric_key_prefix}_{key}"] * _eval_cnt) / (self._agg_metrics['eval_count'] + _eval_cnt)
                    self._agg_metrics['eval_count'] += _eval_cnt
                    metrics[f"eval_{key}"] = self._agg_metrics[key]
                    metrics[f"{metric_key_prefix}_count"] = _eval_cnt
                    metrics[f"eval_agg_count"] = self._agg_metrics['eval_count']

        all_labels = None
        all_preds = None
        num_samples = _eval_cnt
        self._metric_agg_idx = (self._metric_agg_idx + 1) % self._num_evals
        return EvalLoopOutput(predictions=all_preds, label_ids=all_labels, metrics=metrics, num_samples=num_samples)
    
if __name__ == "__main__":
    trainer = GenerateEvalSFTTrainer()



#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
from proof_wala.llm_helpers.llama_chatbot import LlamaConsoleChatBot

def run_chat_bot(bot_name: str, model_name: str, system_prompt_filepath: str, sample_conversation_filepath: str):
    bot = LlamaConsoleChatBot(bot_name, model_name, system_prompt_filepath, sample_conversation_filepath)
    bot.load()
    bot.run()
    pass


#!/usr/bin/env python3

import sys
root_dir = f"{__file__.split('proof_wala')[0]}"
if root_dir not in sys.path:
    sys.path.append(root_dir)
import typing
import copy
from proof_wala.llm_helpers.model import TrainingDataFormatterCallback
from proof_wala.llm_helpers.theorem_proving_training_dataset import TheoremProvingTrainingDataset
from itp_interface.tools.training_data import TrainingData
from proof_wala.parsers.copra_grammars import CoqGPTResponseDfsGrammar, CoqGPTRequestGrammar, CoqGptRequest, CoqGptResponse
from proof_wala.parsers.grammars.prompt_template_grammar import PromptGrammar

class CodeT5TrainingDataset(TheoremProvingTrainingDataset):
    response_grammar: CoqGPTResponseDfsGrammar = CoqGPTResponseDfsGrammar()
    request_grammar: CoqGPTRequestGrammar = CoqGPTRequestGrammar()
    def __init__(self, 
            training_data: TrainingData,
            characters_per_token: float = 3.6,
            max_tokens: int = 2048,
            no_steps: bool = False,
            lang_tag: typing.Optional[str] = None):
        assert characters_per_token >= 1, f"Characters per token must be at least 1, but is {characters_per_token}"
        super().__init__(training_data)
        self.response_grammar = CoqGPTResponseDfsGrammar()
        self.request_grammar = CoqGPTRequestGrammar()
        self.prompt_grammar = PromptGrammar()
        self.characters_per_token = characters_per_token
        self.max_tokens = max_tokens
        self.max_tokens_in_prompt = self.max_tokens
        self.max_chars_in_prompt = int(self.max_tokens_in_prompt * self.characters_per_token)
        self.no_steps = no_steps
        self.lang_tag = lang_tag
        if self.max_tokens_in_prompt < 0:
            raise ValueError(f"Max tokens in prompt is negative: {self.max_tokens_in_prompt}, increase max tokens or decrease characters per token")
        self.prompt_delimiter = "\n"

    def __getitem__(self, idx):
        response, request = self._get_response_request(idx)
        prompt = self.response_grammar.format_as_per_grammar(response, max_token_cnt=self.max_tokens_in_prompt, characters_per_token=self.characters_per_token)
        completion = self.request_grammar.generate_message_from_gpt_request(request)
        prompt += self.prompt_delimiter
        if self.lang_tag is not None:
            prompt = f"{self.lang_tag}\n{prompt}"
        return {
            "prompt": prompt,
            "completion": completion
        }

    def _get_response_request(self, idx):
        steps : typing.List[str] = []
        last_step = None
        last_tdf = self.training_data[idx]
        if not self.no_steps:
            assert last_tdf.proof_id is not None, f"Proof id is None for training data at index {idx}, cannot infer the previous steps in the proof"
            tdf = self.training_data[idx - 1] if idx > 0 else None
            last_step = tdf.proof_steps if tdf is not None and tdf.proof_id == last_tdf.proof_id else None
            tdf = self.training_data[idx - 2] if idx > 1 else None
            while tdf is not None and last_tdf.proof_id == tdf.proof_id:
                steps.extend(reversed(tdf.proof_steps))
                idx -= 1
                tdf = self.training_data[idx - 1] if idx > 1 else None
            steps.reverse()
        response = CoqGptResponse(
            success=True,
            steps=steps,
            last_step='\n'.join(last_step) if last_step is not None else None,
            training_data_format=last_tdf
        )
        request = CoqGptRequest(
            args=last_tdf.proof_steps
        )
        return response, request
    
    def prompt_formatter(
            response: CoqGptResponse, 
            max_tokens_in_prompt: int = None, 
            characters_per_token: float = 4,
            no_steps: bool = False,
            lang_tag: typing.Optional[str] = None) -> str:
        response_copy = copy.deepcopy(response)
        if no_steps:
            response_copy.steps = []
            response_copy.last_step = None
        prompt = CodeT5TrainingDataset.response_grammar.format_as_per_grammar(
            response_copy,
            max_token_cnt=max_tokens_in_prompt, 
            characters_per_token=characters_per_token)
        if lang_tag is not None:
            prompt = f"{lang_tag}\n{prompt}"
        return prompt

    def response_parser(request_str: str) -> CoqGptRequest:
        if request_str.endswith(CodeT5PromptTrainingDataFormatter.stopping_token):
            request_str = request_str[:-len(CodeT5PromptTrainingDataFormatter.stopping_token)]
        request, _ = CodeT5TrainingDataset.request_grammar.attempt_parsing(request_str)
        return request
    

class CodeT5PromptTrainingDataFormatter(TrainingDataFormatterCallback):
    stopping_token = "[END]"
    def __call__(self, training_data_format_examples: typing.List[typing.Dict[str, str]]) -> typing.List[str]:
        output_texts = []
        all_prompts = training_data_format_examples["prompt"]
        all_completions = training_data_format_examples["completion"]
        stopping_token = ""
        for prompt, completion in zip(all_prompts, all_completions):
            output_text = f"{prompt}{completion}{stopping_token}"
            output_texts.append(output_text)
        return output_texts
    
    def get_prompt_and_completion(self, training_data_examples: typing.List[typing.Dict[str, str]]) -> typing.List[typing.Tuple[str, str]]:
        all_prompts = training_data_examples["prompt"]
        all_completions = training_data_examples["completion"]
        prompt_and_completions = []
        if isinstance(all_prompts, str):
            all_prompts = [all_prompts]
            assert isinstance(all_completions, str)
            all_completions = [all_completions]
        assert len(all_prompts) == len(all_completions)
        for prompt, completion in zip(all_prompts, all_completions):
            # Trim everything after the stopping token from the completion
            completion = completion.split(self.get_stopping_token())[0]
            prompt_and_completion = (prompt, completion + self.get_stopping_token())
            prompt_and_completions.append(prompt_and_completion)
        return prompt_and_completions
    
    def get_stopping_token(self):
        return CodeT5PromptTrainingDataFormatter.stopping_token

if __name__ == "__main__":
    prompt_name = "copra-coq-dfs"
    data_folder = f".log/traces"
    meta_filename = "local.meta.json"
    training_data = TrainingData(data_folder, meta_filename)
    # Not removing system_prompt_file and conversation_prompt_file though not needed
    # as removing makes the function signature incompatable with the call-type in run_training.py
    with CodeT5TrainingDataset(training_data, max_tokens=4096, no_steps=True) as dataset:
        hf_dataset = dataset.get_hf_dataset()
        formatter = CodeT5PromptTrainingDataFormatter()
        formatted_dataset = formatter(hf_dataset)
        for example in hf_dataset:
            prompt = example["prompt"]
            completion = example["completion"]
            print(f"Prompt:\n{prompt}")
            print(f"Completion:\n{completion}")
            print('-'*80)
        prompt_and_completions = formatter.get_prompt_and_completion(hf_dataset)
    pass